/*
    base and operators defining the Hilbert space
    base = spin-configurations,
    p.e. get the index from config or reverse, get symetric conf.
    - look at block_no
  
 */
#include "hilbert.h"
#include "mem.h"
#include "tbase.h"
#include "ansatz.h"
#include "array_m.h"
#include "block.h"       /* mpi_* declarations */
/* open,lseek,close using off_t (64bit using LFS) */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif
#ifdef CONFIG_MPI
#include <mpi.h>
#endif

ULong  npa[NN+1][NN+1][NN+1];   /* (o+u+d)!/(o!u!d!) stored as 3D table */

zahl fak(int i){zahl r=1.0;int j;for(j=2;j<=i;j++)r*=j;return(r);}
/* ------------------------------------------------------
 *  calculate the binominal coefficient (use pascal triangle(?))
 *  over(n,k) = n!/((n-k)!k!) = over(n-1,k-1)*n/k = over(n-1,k)+over(n-1,r-1)
 * --------------------------------------------------------- */
ULong over(int n,int k){
  static ULong pascal[2*NN+1][2*NN+1]; static int ok=0; int i,j;
  if(!ok){ ok=1;		/* de: Pascal-Dreieck bauen */
    for(i=0;i<=2*NN;i++)for(j=0;j<=2*NN;j++)pascal[i][j]=(ULong)((j)?0l:1l);
    for(i=1;i<=2*NN;i++)for(j=1;j<=i;j++)
      pascal[i][j]=pascal[i-1][j]+pascal[i-1][j-1];	/* OVL ? */
  }  
  if(n>2*NN || k>n) return 0l;		/* NN => 2*NN  Nov98 JS */ 
  return pascal[n][k]; /* only 0<=k<=n<=NN */    
}


#if CM(1,1,0)  
/* Perm. mit Wdhlg (3 Elemente)=(o+u+d)!/(o!u!d!)=over(n-u,d)*over(n,u) */
 ULong np(int o,int u,int d){ return(over(o+u+d,u)*over(o+d,d));}
#else	/* Produkt 2er Kombinat. (2 Elemente)=(n over u)(n over d) */
 ULong np(int n,int u,int d){return(over(n,u)*over(n,d));}
#endif
void   paini(){ int a,b,c;
    for(a=0;a<=NN;a++)for(b=0;b<=NN;b++)
    for(c=0;c<=NN;c++)npa[a][b][c]=np(a,b,c);
    return;}
    
/* number of spin configurations (nn,nu,nd), same as n1 without Sym */
ULong numconf(){ return(npa[nn-CM(1,1,0)*(nu+nd)][nu][nd]);}

/* --------------------------------------------------------------------*
 *                    in-out-functions
 * call this function not more than 8 times within one printf() :) 
 *  v2.26:  .=empty-site, u=up, d=down, 3=up+down 
 *  v2.27:  .=empty-site, o=up, x=down, Q=up+down (looks better)
 *          think on pointers pointing to you "o" or away from you "x"
 * ToDo: for S>1/2 output for tJ,tU (s,m=-s...+s) and for JJ (m)=(-s...+s)
 * - not thread safe
 * ------------------------------------------------------------------- */
char *ccfg(tbase ic){ /* Spinconfiguration |543210> zB: |o.xoxo> */
  static char s1[8*(NN+1)], s2[4+16+1]=".oxQ0123456789abcdef";
  char *s3; static int x=0; int i;
  x=(x+1)%8; s3=s1+x*(NN+1);  /* NN+1 Bytes per String, max. 8 strings */
#if 0 // old site wise
  for (i=nn-1;i>=0;i--) s3[nn-1-i]=s2[ geti(ic,i) ];
  s3[nn]='\0';
#else  /* 2012-06-27 */
  s3[0]='0'; s3[1]='x';
  for (i=(nn*CM(1,2,2)+3)/4-1;i>=0;i--)
    s3[2+(nn*CM(1,2,2)+3)/4-1-i]= s2[(15&(ic>>(4*i)))+4];
  s3[  2+(nn*CM(1,2,2)+3)/4]='\0';
#endif
  return(s3);
}

/* ----------------------------------------------------------------------- * 
 * symmetry-OP   PermOp(ip)*ib1=s*ib2 , s = +1/-1  ip = permutation vector 
 * ----------------------------------------------------------------------- */
/*   example:
 P = [0 2 1 3] = (0) (1,2) (3)
 ib1= (.. du .u d.) = sites (3 2 1 0) = bits (76 54 32 10)
 m1=7=3d -
 m1=6=3u -						 
 m1=5=2d m2=3=1d i2=(.. .. d. ..) s=0       i3=(11 11 10 00)=(0000,1110)
 m1=4=2u m2=2=1u i2=(.. .. du ..) s^=i3_1=0 i3=(00 00 01 00)=(1110,1110)
 m1=3=1d -
 m1=2=1u m2=4=2u i2=(.. .u du ..) s^=i3_3=0 i3=(11 11 01 00)=(0010,1110)
 m1=1=0d m2=1=0d i2=(.. .u du d.) s^=i3_0=0 i3=(00 00 10 10)=(0010,0001)
 m1=0=0u -      
 ------------------------------------------------------------------------ */
/* ms!=0 multiply phase with k, ToDo: remove and simplify
 * later build in FPGA!
 */
/* static inline */ /* check asm-code for inlining gcc -S -dA */
/* tphase is complex in complex mode, ToDo: always signed char? +/-1 */
void get_symconfig(tbase ib1,int ip,tbase *ib2,tphase *ss,int ms){
  register tbase i2;
  register int m1, m2;
#if CM(0,1,1)
  register tbase i3; int s;
#endif
  *ib2=ib1; *ss=+1; if(ip<0) return; /* 1-Operator (0,identity) */
#if CM(0,1,1)
  // printf("\n symcfg_i= %s %d",ccfg(ib1),ip);
  // permutation per creation/annihilation: ciu.cid.aiu.aid ???
  // from:  uX.   dX.  3X.  dXu        3Xu   3Xd  see utils/fermion_sign.c
  // to:  +p.Xu +p.Xd +.X3 -uXd      +puX3 +pdX3   p=Parity(X)
  i3=i2=0l; s=0;			    /* remove from NN-1 to 0 */ 
  for(m1=2*nn-1;m1>=0;m1--)if((ib1>>m1)&1){ /* m1 = old index */
    m2 = sym.i[ip][m1/2]*2  | (m1&1);	    /* m2 = new index */
    i2 |= ((tbase)1) << m2;		    /* move spin to site m2 */
    if (m2>0) s^=(int)(1&(i3>>(m2-1)));     /* number of spins jumped over */ 
    i3 ^= (~(tbase)0)<<m2;		/* derived fermionic sign, 64bit?? */
  }
  // printf("\n symcfg_o= %s %d",ccfg(i2),s);
  (*ss) *= 1-2*(s&1);   /* ToDo: return s in units of I*2Pi/2 */
#else /* CM(1,0,0) Heisenberg, ToDo: make speedtests over all n1 */ 
  i2=0l;
#if 1  /* how much asm commands? */
  for (m1=0; ib1!=0 && m1<nn; m1++, ib1>>=1) if (1&ib1) { /* faster if nd<=nu */
    m2 = sym.i[ip][m1];
    i2 |= ((tbase)1) << m2; /* ToDo: m2>32 failes on 32bit gcc! use table[NN] */
  }	/* ToDo: uebergabe &sym.i[ip] statt ip ???  */
#else
  t_sites *a=sym.i[ip];
  for (m1=NN; ; ) { /* faster if nd<=nu */
    i2 |= ((tbase)(1&ib1)) << *a;      /* asm: jmp if NN>sizeof(int) */
    a++; m1--; ib1>>=1; if (!m1 || !ib1) break;
  }	/* ToDo: uebergabe &sym.i[ip] statt ip ???  */
#endif
#endif /* JJ */
  *ib2=i2;
  if(ms){
    (*ss)*=sym.w[ip];   /* better add fractional angles (integer) outside */
#if VecType & 4 /* complex phase */
    /* remove rounding errors, do it outside */
    if(fabs(imag(*ss))<1e-9) (*ss)=C(real(*ss),0);
#endif
  }  
  return;
}

tbase m55=0;    /* dynamic initialisation only in C++ => initmasks */
tbase mAA=0;
tbase m0F=0;

/* initialize bitmasks m55,mAA,m0F */
int initmasks(int nn){
  int i; tbase r=0;
  for(i=0;(unsigned)i<sizeof(tbase)*8;i+=2) r|=((tbase)1)<<i;
  m55=r;
  mAA=r<<1;
  m0F=~(((~(tbase)0)<<(nn-1))<<1); /* (int32)<<32 won't work on x86 */
  /*  printf(" ini55: %s\n",ccfg(r)); */
  return 0;
}
    

/* --------------------------------------------------------------------	*
 * 		Spin-Konfigurationen ki und deren Index i		*
 * --------------------------------------------------------------------	*/
#if NN == 8*8 /* JS2012-03-06 fix 64bit overflow, but slower, x=0 never happens? */
#define MF(x)   (((x)==(NN))?~(tbase)0:~((~(tbase)  0)<<(x))) /* x=0..64 */
#else
#define MF(x)   ( ~((~(tbase)0)<<(x)))	/* << only int/long, not longlong */
#endif
#define MA(x)   (mAA & MF(x))		/* MF(x) = x * 1 bits */
#define M5(x)   (m55 & MF(x))
#define M1(x)   (  ((tbase)  1)<<(x) )  /* undefined for x>=64 */
#define M2(x)   (  ((tbase)  2)<<(x) )
#define M3(x)   (  ((tbase)  3)<<(x) )
#define ME(x)   (  ((tbase)0-2)<<(x) )
#define MC(x)   (  ((tbase)0-4)<<(x) )

// works until 64 bits, JS 2012-05-31 (faster than loop)
static inline int ctz64(tbase v){ // count trailing zero bits using bintree search
#if NN <= CM(32,16,16)
  return __builtin_ctz(int(v));
#else
  unsigned int c;
  if (v & 1) return  0;
  if (!v)    return 64;
  c=1;
#if    NN > CM(32,16,16)
  if ((v & 0xffffffff) == 0) { v >>= 32; c += 32; }
#endif
  if ((v & 0xffff    ) == 0) { v >>= 16; c += 16; }
  if ((v & 0xff      ) == 0) { v >>= 8;  c += 8;  }
  if ((v & 0xf       ) == 0) { v >>= 4;  c += 4;  }
  if ((v & 0x3       ) == 0) { v >>= 2;  c += 2;  }
  c -= v & 0x1;
  return c; // Warning 0 returns 31
#endif
}

#if CM(0,1,0)
inline tbase next(tbase cfg){                                       /* ki++ */ 
  int i=0,d=0,u=0; 	/* Achtung: ki<<32 bei 32bit nicht definiert*/
  if(cfg&2l)d+=2; else if(cfg&1l)u+=2;
  while((i+=2)<2*nn)switch((int)((cfg>>i)&3l)){
   case 0:       if(u){ u-=2;
     return((cfg & MC(i)) | M1(i) | M5(u)<<d | MA(d));} else if(d){ d-=2;
     return((cfg & MC(i)) | M2(i) | M5(u)<<d | MA(d));} break;
   case 1: u+=2; if(d){ d-=2;
     return((cfg & MC(i)) | M2(i) | M5(u)<<d | MA(d));} break;
   case 2: d+=2;}
  return(M5(u)<<d | MA(d));}

tbase  ll1(ULong ii){ /* getbasis from index 0..N1-1 */
  ULong i,t;tbase j;int io,iu,id;
  j=0l;io=nn-nu-nd;iu=nu;id=nd;i=ii+1l;
  for(;;){
   if(io>0){ if(i<=(t=npa[io-1][iu][id])){j=(j<<2);  io--;continue;}i-=t;}
   if(iu>0){ if(i<=(t=npa[io][iu-1][id])){j=(j<<2)|1;iu--;continue;}i-=t;}
   if(id>0){ if(i<=(t=npa[io][iu][id-1])){j=(j<<2)|2;id--;continue;}i-=t;}
   else break;}
  return(j);}
    
ULong ll2(tbase ii){ /* n-dimensionale Linkodierung */
/* Inp: Spinconfiguration, Out: Index des Basiszustandes */
    ULong j=0l;int io,iu,id,i,i1;
    io=nn-nu-nd;iu=nu;id=nd;for(i=nn-1;i>=0;i--){
      i1=(int)(3&(ii>>(2*i)));
      if(i1>0 && io>0)j+=npa[io-1][iu][id];
      if(i1>1 && iu>0)j+=npa[io][iu-1][id];
      if(i1==0)io--;
      if(i1==1)iu--;
      if(i1==2)id--;}
    return(j);}

#elif CM(1,0,0)

/* 01101 i=1 d=1 inp  */
/* 01110 i=2 d=1 !!!  */
/* 10011 i=2 d=2      */
/* 10101 i=1 d=1      */
/* 10110 i=2 d=1      */
/* 12233=123=LM
   00011 002 i=2 d=1
   00101 011
   01100 020
   10001 101
   10100 110
 */
/* last change: long long Anpassung 9.4.1998 JS */
/* MP-version Dec99, better readable */
/* improved next()-function, some easy symmetries included */
/* 2012-03-06 64bit trouble for NN=64 */
static inline tbase next(tbase cfg){	/* JJ */ 
  int i,d;     /* i - position of first "01"-bit sequence from bit0 */
               /* d - number of 1-bits exclusive 01 pattern */
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){   /* ToDo: replace by S(+,i)-ops and s>1/2 sites */
    int k,j;
    static tbase ib_last=0;
    i=d=0;
    // input pattern must have LM-symmetry! 2012-05-16
    // printf("\nDBG0 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    // for non-s1sym-input output may also be non-s1sym
    // aaa aaa aaa bbb bbb bbb 9*s2/2 (triangle3003)
    // aaa bbb ccc ddd eee fff 3*s6/2
    // 100 000 000 011 111 111 is a invalid entry! ("10" on most left site)
    // 100 000 000 101 111 111 next valid s2/2
    // 100 100 100 100 101 111 next valid s6/2
// re-idx= 0 1   2  3  4  5   6  7  8  9  10 11 12 13  14 15 16 17 size(sym)=5.27
// LS=3   6  6   6  6  6  6   6  6  6  6   6  6  6  6   6  6  6  6
// LI=3   0  0   0  1  1  1   2  2  2  3   3  3  4  4   4  5  5  5
// LM=3   0  1   2  1  2  0   2  0  1  0   1  2  1  2   0  2  0  1 !!!.lm
// 011675  . 1   . -.- .  1   .  1  1  .   .  1  1  1   0  1  0  1 =noSymStart
// 0156a3  . 1   .  1  .  1   .  1  1  .   1  .  1  .   .  .  1  1 = bad_next
// 0142af  . 1   .  1  .  .   .  .  1  .   1  .  1  .   1  1  1  1 =good_next
// LN=3   5  3   4  8  6  7  11  9 10 14  12 13 17 15  16 18 18 18 next_s1site
// L0=    0  1   2  1  2  0 ... startidx or preidx
// 0201bf  1 .   .  .  .  .   .  .  .  1   1  .  1  1   1  1  1  1 =noSymStart
// 0215ab  1 .   .  .  .  1   .  1  .  1   1  .  1  .   1  .  1  1 = bad_next
// 02151f  1 .   .  .  .  1   .  1  .  1   .  .  .  1   1  1  1  1=good_next
// 0453f   . .   .  1  .  .   .  1  .  1   .  .  1  1   1  1  1  1 =start
// 047ab   . .   .  1  .  .   .  1  1  1   1  .  1  .   1  .  1  1 =good_next
// 04af5   . .   .  1  .  .   1  .  1  .   1  1  1  1   .  1  .  1 =bad_next
// ToDo: use .lnext forfaster run 2012-05-17
    /* spread spins over higher d-set LM-cluster (needed for noS1sym input) */
    // printf("\nDBG1 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    // ToDo: Problem if more LM-sites than down spins
    int badLM=0;
    if (ib_last != cfg) // we only need this slow part for initial ll1()
    for(j=nn-1;j>0;j--)if((cfg>>j)&1)  // 2012-05-16 s2/2 triangle3003
    for(k=sym.lp[j]; k<nn; k=sym.lp[k]) // faster
//    for(k=j-1;k>=0;k--)if(sym.lm[k]==sym.lm[j])  
                           if(!((cfg>>k)&1)){ 
      /* LM was 1..0xx (bad LM-sym), remove and count all 1-bits behind 0 */
         int i1;
         for(i1=k-1;i1>=0;i1--)if(sym.lm[k]!=sym.lm[i1])
           if((cfg>>i1)&1){d++;cfg&=~M1(i1);}
         if (d==0) { /* not enough down-bits */ j=0;i=0;badLM=0;break;}
         d--;cfg|=M1(k); i=d; badLM=1;
//     printf("DBGx i=%2d d=%2d next=%s\n",i,d,ccfg(cfg));
     // i!=0: bad LM cfg
    }
    if (!badLM) // good LM cfg
    {
      for(;i!=nn;i++){ /* searching 01 bit pattern */
       if(1&(cfg>>i)){ /* clear spin, if no higher LM-spin */
         for(k=sym.ln[i];k<nn;k=sym.ln[k])if(1&(cfg>>k))break; // faster
         if (k==nn)
         for(k=i;k<=i;k=sym.lp[k]) // faster
           if (1&(cfg>>k)) { d++;cfg&=~M1(k); }
       }
       else { /* OVL to other LM */
         for(j=0,k=sym.lp[i];k<i;k=sym.lp[k])j++; // faster
         if (d>j) break;
       }
      }
      if (i<nn) { cfg |= M1(i); d--; } else i=d;
    }
    /* first spread spins over lowerbit sites of same LM-cluster */
    // printf("\nDBG2 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    for (k=sym.lp[i];k<nn && d>0;k=sym.lp[k]) /* 2012-05-24 speedup */
      if (!((cfg>>k)&1)) { d--;cfg|=M1(k); }
    /* now spread over the rest */
    // printf("\nDBG3 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    for(k=0;/* k<i &&*/ d>0;k++)if(!((cfg>>k)&1)){ d--;cfg|=M1(k); }
    // printf("\nDBG4 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    ib_last= cfg; // for skip check
    return(cfg);
  } // S1_SYM
#endif
  /* searching xxxxx0111...11100...0 */
  /*                ^ i=d+1        ToDo: __builtin_ctzll()  */
  // count number of trailing zero bits (i-d) and 1-bits (d) 
  // gcc-3.2 __builtin_ctz x86 (add fkt for vlint.h)  ctz=ffs-1
  // http://en.wikipedia.org/wiki/Find_first_set
  //  (12)ooxoxoOxxxoo(0)  i=5 d=3 5-loops (2 trailing zeros)
  //  (12)ooxoxoXOOOXX(0)   i=(d)?d:i+1...
#if NN<=64 && 1 // triang 6 0 0 9 8+46 0sym n1=1040e6 156s  46+8=3.63m
  i=ctz64(cfg); d=ctz64(~(cfg>>i)); i+=d; // count trailing zero bits
#else           // triang 6 0 0 9 8+46 0sym n1=1040e6 159s  46+8=3.63m
  i=d=0;for(;i!=nn;i++){ if((cfg>>i)&1)d++;else if(d)break; }
#endif
  if (i==nn) return MF(d);  /* JS 2012-03-06 fix overrun */
  d--;                        /* one d-bit is shifted higher, d=rest */
  // clear i bits + set 1 bit + set d lowest bits
  return ( (cfg & ~MF(i)) | M1(i) | MF(d) ); /* d==0: x<<-1 ???  */
}

tbase  ll1(ULong ii){	/* getbasis from index 0..N1-1 (no-sym) */
  ULong i,t;tbase j;int iu,id;
  j=0l; iu=nu; id=nd; i=ii+1l; /* j overflow if ii>=nns && nn=64 */
  if(nn!=nu+nd)F0("nu+nd != nn");
  for(;;){
    if(iu>0){ if(i<=(t=npa[0][iu-1][id])){j=(j<<1)|0;iu--;continue;}i-=t;}
    /* else { j=(j<<id)|MF(id);break; F A S T E R ???? } */
    if(id>0){ if(i<=(t=npa[0][iu][id-1])){j=(j<<1)|1;id--;continue;}i-=t;}
    break;}
  return(j);
}
    
ULong ll2(tbase ii){	/* n-dimensional Lin-coding */
/* Inp: spin configuration, Out: index of base state */
  ULong j=0l; int iu,id,i,i1;
  iu=nu; id=nd; if(nn!=nu+nd) F0("nu+nd != NN")
  for(i=nn-1;i>=0;i--){
    i1=(int)(1&(ii>>i));
    if (i1>0 && iu>0 && id>0) j+=npa[0][iu-1][id];
    if (i1==0) iu--;
    if (i1==1) id--;}
  return(j);
}

#elif CM(0,0,1)  /* now the coding for the Hubbard model */

inline tbase next(tbase cfg){  				/* ki++ */ 
  int i,d,u;if(cfg&1l)u=2;else u=0;if(cfg&2l)d=2;else d=0;  
  for(i=2;i<2*nn;i+=2)switch((int)(3&(cfg>>i))){      /* -1<<2=-4 */
   case 0:
    if(u){ return((cfg & MC(i)) | M1(i) | M5(u-2) | MA(d));} else 
    if(d){ return((cfg & MC(i)) | M2(i) | M5(u) | MA(d-2));} break;
   case 1:  u+=2; 
    if(d){ return((cfg & MC(i)) | M2(i) | M5(u) | MA(d-2));} break;
   case 2:
    if(u){ return((cfg & MC(i)) | M3(i) | M5(u-2) | MA(d));} d+=2; break;
   case 3:  u+=2;d+=2;}
  return( M5(u) | MA(d) );}

tbase  ll1(ULong ii){	/* get configuration from index 0..N1-1 */
    tbase j;int in,iu,id;ULong i,t;i=ii;j=0l;iu=nu;id=nd;
    for(in=nn-1;in>=0;in--){
     if(in>=iu && in>=id){
         if(i<(t=npa[in][iu  ][id  ])){continue;}i-=t;}
     if(iu>0 && in>=id){
         if(i<(t=npa[in][iu-1][id  ])){j|=M1(2*in);iu--;continue;}i-=t;}
     if(id>0 && in>=iu){
         if(i<(t=npa[in][iu  ][id-1])){j|=M2(2*in);id--;continue;}i-=t;}
     if(id>0 && iu>0)
         if(i<   npa[in][iu-1][id-1] ){j|=M3(2*in);id--;iu--;continue;}
     break;}
    return(j);}
    
ULong ll2(tbase ii){                /* ki -> i */
    ULong j;int iu,id,i,i1;j=0l;iu=nu;id=nd;
    for(i=nn-1;i;i--){ if(i<iu)if(i<id)break;i1=3&(ii>>(i<<1));
      if(i1>0){  if(i>=iu   && i>=id        )j+=npa[i][iu  ][id];
       if(i1>1){ if(i>=iu-1 && i>=id && iu>0)j+=npa[i][iu-1][id];id--;
        if(i1>2){if(i>=iu   && i>=id        )j+=npa[i][iu  ][id];iu--;}}
       else iu--;}}
    return(j);}
#endif
#undef MF
#undef MA
#undef M5
#undef M1
#undef M2
#undef M3
#undef MC
#undef ME

/* suche entry in sorted tab[0..l-1], not found => 0 */
/* ToDo: it should be easier with uuu and ddd-tab if tU */
ULong findcfg(tbase entry, ULong l){
  ULong j,m;j=0l;  /* v203 tbase1=>ULong, 98: ULong=>tbase1 */
#if (CM(1,2,2)*NN-1)>8*S_long-1
  m=~(~(ULong)0u>>1);		/* set msb (highest bit) */
#else
  m=((ULong)1)<<(CM(1,2,2)*nn-1);
  if(m==0 || m==1)F0("OVL");
#endif
  while(m){if((j|m)<l)	/* long unsigned == unsigned long ? */
    if(getl1(j|m)<=entry) j|=m;
    m>>=1;
  }
  /* if(getl1(j)!=entry)j=0l;			 not found */
  return((ULong) j);			/* ULong/tbase1 ??? */
}

char *outphase(tphase r){  /* tphase: complex float or integer */
  static char s1[100];
#if VecType & 4
  double phase=arg(r)/M_PI; if (phase<0) phase+=2;
  /* sprintf(s1,"+(% g,% g)",real(r),imag(r)); */
  sprintf(s1,"(1%+5.0e)*phase(%5.3f)", fabs(r)-1, phase); // JS2012 1-r to r-1
#else
  sprintf(s1,"%+d",r); 
#endif
  return(s1);}


/* ------------------------ symmetric base config (class Hbase) ----
 * inp: sym.k[], sym.i[] 
 * out: |rc)=sign(s)*|s|^(-1/2)*|rc],  s=<ib1|rc)
 * |] is sym-base, summe aller zu |ib1> symmetr. configs 
 * |) is normed sym-base, 
 * |> is spin-configuration (1st base) 
 * ------------------------------------------------------------- */
 
/*
 *  print symmetric base config for debugging purposes
 *  $len>0 limits the output to $len lines
 */
int b_out(struct sbase *b,int len){
  int i;if(len==0)len=b->num;
  printf("\n sign=%s norm2=%4g num=%4d o=%d scfg=%s",
    outphase(b->sign),(double)b->norm2,b->num,b->o,ccfg(b->scfg));
  for(i=0;i<len && i<b->num;i++) 
    printf("\n%3d %s %s",i,outphase(b->is[i]),ccfg(b->ib[i]));
  return 0;
}

#if VecType & 4		/* used in b_getbase... */
#define ABS1(x) ((int)(fabs(x)+.5))
#else
#define ABS1(x) abs(x)
#endif

/* --- ax|vx>+...+1|vx> => (ax+1)|vx>+...+0|vx> --- ib[num] already stored
 * replace 0-entries by (b.num-1)-entries and b.num-- (pack) ???? 
 * the same cfg-entries with different signs do interfere to 0! 
 * twice=0: the caller function knows the twice factor already
 *     we dont have to recalculate it (ToDo: check!)
 * called by b_getbase()
 */
void b_twice(struct sbase *b,int twice){  
  int k,j,num/* ,t */;
  if(twice==0)
  for(num=0;num<b->num;num++) if (b->ib[num]==b->scfg) twice++;
  if(b->o!=2 && twice>1)
  for(num=0;num<b->num;num++) if (ABS1(b->is[num])) {
    for(k=num+1,j=twice;k<b->num && j>0;) if (b->ib[k]==b->ib[num]) {
      b->num--;j--;
      b->is[k]=b->is[b->num];		/* reduce table size */
      b->ib[k]=b->ib[b->num];
    } else k++; /* Jan03: "else" correct? new {} */
  }
  b->norm2=b->num;
  return;
}

/* ----------------------------------------------------------------------- * 
 * Sz-symmetry (spin inversion) u<->d,   attention: ss may point to s1 !!!
 *  ToDo: change type of tphase from complex value (k*i*2pi/l) to integer angle (k)
 *  ToDo: dont input s1, better return the sign
 *  used in xval.c wop_ud() if nu==nd obsoletem and in hilbert.c
 * ----------------------------------------------------------------------- */
static inline  // replace tphase by sig-bit (char/int)
void get_sudconfig(tbase ib1,tphase s1,tbase *ib2,tphase *ss){
  /* ib2(ss) can also point to ib1(s1) !!! */
  *ib2=ib1;*ss=s1;		/* default: 1-operator */
  if(nu!=nd || !sym.wud) return; /* ToDo: remove it, decide it outside! */
#if CM(1,0,0)
  (*ib2)=ib1 ^ m0F;
#else /* tJ + tU */
  (*ib2)=((mAA & ib1)>>1) | ((m55 & ib1)<<1); // ToDo: ((ib1>>1)&M55)|...
#if CM(0,0,1)  /* 3=(u+d) will change sign, ToDo: only if sign needed? slow */
  { tbase1 i3; // ToDo2012: remove if (bit hacks) + logcount ib1&(ib1/2)&m55
     // better rightshift ib1 until zero? &3 
    for (i3=((tbase)3)<<(2*nn-2); i3!=0; i3>>=2) if ((ib1&i3)==i3) (*ss)*=-1;
  }
#endif /* tU only */
#endif /* tJ + tU */
  (*ss)*=sym.wud;	/* ToDo: can be done outside, return only +/-1 */
  return;
}

/* ---- get_lm_factor ---- (long64: 2^64=1.845e19)
 * N=6 s=14 s*N=84 sym=12 spin-chain, J1=J2=1
 *   u=18,d=66: ns=1e18 n1= 2888 n2max=2e16 E0=-15   2s,8s (fits to 64bit)
 *   u=20,d=64: ns=1e19 n1= 4440 n2max=1e17 E0=-62         (ERR9112 k=0)
 *   u=22,d=62: ns=9e19 n1= 6500 n2max=1e18 E0=-105  3s,19s
 *   u=24,d=60: ns=2e19 n1= 9097 n2max=9e18 E0=-144  4s,26s
 *   u=26,d=58: ns=1e19 n1=12208 n2max=2e19 E0=-179  5s,35s
 *   u=42,d=42: ns=2e19 n1=17052 n2max=2e22 E0=-309 i70=169s
 * Tnorm2 is ULong per default, but can be set to double for big-s systems
 */
Tnorm2 b_factor_lm(tbase ib1){  
  int k,b1,i1,obj[NN][4]; /* 0,1,2,3 for empty,up,down,up+down sites */
  Tnorm2 n2, factor;
  n2=1;	
  if(sym.num_lm==nn) return(n2);
  /* compute factor (n1+n2+n3+n4)!/(n1!n2!n3!n4!) =
   *   = (n1+n2+n3+n4)!/n1! * (n2+n3+n4)!/n2! * (n3+n4)!/n3!
   *  where n1=number_of_empty_sites, n2=number_of_ups, ... per LM-spin
   */
  for(k=0;k<nn;k++){ obj[k][0]=obj[k][1]=obj[k][2]=obj[k][3]=0; } 
  for(i1=0;i1<nn;i1++){  /* count types of sites n1,..,n4 */
      b1=geti(ib1,i1); obj[sym.lm[i1]][b1]++;
  }
  for(k=0;k<nn;k++)if(sym.lm[k]==k){
      factor=over(obj[k][0]+obj[k][1]+obj[k][2]+obj[k][3],obj[k][3])
            *over(obj[k][0]+obj[k][1]+obj[k][2]          ,obj[k][2])
            *over(obj[k][0]+obj[k][1]                    ,obj[k][1]);
      if( (double)(n2*factor) != n2*(double)factor ) ERR(9110);
      n2*=factor;
  }
  return(n2);
}

/* ----------------------------------------------------------------------- */
#define Err17 {static int e0=1;if(e0){ E0("b_*_lm may be buggy! check with NOS1SYM!"); e0=0;}}

/*  effect of 2-site-LM-symmetry for tU-model considering fermionic sign (Feb04)
 *      .. => +..              .3 => +3.           33 => +33
 *      .u => +u.*parity       ud => -du
 *      uu => -uu vanished     u3 => +3u*parity
 *   LM = c(+,p[i])c(+,p[j])c(-,j)c(-,i)              */
/* Feb04: rewritten
   ToDo: test S=5/2 N=6 16+14 nk=1.5e8 sym=4 n1=1091 33s until 2S=
       does not work for LM, tU u2d0 N=8 square
   oo=0: return full expanded base (all symmetric configs)
   oo=1: return smallest + sign + norm2 (dont extend base)
   oo=2: break, if sign==0 or smaller found (like ifsmallest)
   ToDo: better return was_smallest/got_smallest/0?
 */
tbase b_getbase_lm(struct sbase *b, int oo){  
  tbase ib0,ib2; int j,b1,b2,i1,i2,n2; tphase s0,s1,s2;
  CM({},Err17,Err17)	/* signum ??? */
  n2=0;	/* count same lm-scfgs */
  /* b->num must be greater than 0 (one argument or more) */
  b->sign=b->is[0]; b->scfg=b->ib[0]; /* if we have n2==1 */
  for(j=0;(ULong)j<(ULong)b->num;j++){
   ib0=b->ib[j]; s0=s1=b->is[j];	/* check next config */
   if(s0==0) E0("Warning: sign[j]=0");
   if (ib0< b->scfg) { b->scfg=ib0; b->sign=s0; n2=1; }
   /* ib0 = original (smallest?), ib2 = new_lm_cfg */
   for(i1=0;i1<nn-1;i1++)for(i2=i1+1;i2<nn;i2++)if(sym.lm[i1]==sym.lm[i2]) {
     b1=geti(ib0,i1); /* site 1 within LM-site */
     b2=geti(ib0,i2); /* site 2 within LM-site */
     if(b1!=b2){ /* exchange sym-sites */
       ib2=puti(puti(ib0,i1,b2),i2,b1); s2=s1;
       /* tU,tJ fermionic sign change? call (2sites) permutation function? */
#if CM(0,1,1)
       if (zi(ib0,i1)+zi(ib0,i2)!=0) { /* examples: .u u3 (fermionic transfer) */
         s2=s1*parity(ib0 & p_mask[i1][i2]); }  /* not coding independent */
       if (zizj(ib0,i1,i2)!=0){ s2=-s1; }      /* ud (fermionic exchange) */
#endif
       if(b->scfg==ib2){	/* same lm-symmetric config stored */
         if(!phase_eq(s2,b->sign)){
         b->sign=0; b->num=0; b->norm2=0; return TBASE0;}
       }
       if(!(ib0<=ib2)){ b->ib[j]=ib0=ib2; b->is[j]=s0=s2;
         if(!(b->scfg<=ib0)){
           b->scfg=ib0; b->sign=s0; n2=0; if(oo==2) return ib2; }
         i1=i2=nn; j--; /* restart for this config */ break;
       }
     }
#if CM(0,1,1)
     else {  /* b1==b2 */
       if (abs(zi(ib0,i1))>0 /* && abs(zi(ib0,i2))>0 */){
         b->sign=0; b->num=0; b->norm2=0; return TBASE0; } /* uu + -uu */
     }
#endif
   }
   /* add factor_lm new lm-symmetric configs to table *slow!* */
   if (oo==0) { Err17; } /* not supported yet, we should not need it */
  }
  /* calculation of n2 is buggy, so we reat it here */
  for(n2=j=0;(ULong)j<(ULong)b->num;j++){
   if (b->ib[j]==b->scfg) n2++; /* how often is scfg stored within table */
  }
  if(oo<2){	/* (n1+n2+n3+n4)!/(n1!n2!n3!n4!) */
    Tnorm2 factor=b_factor_lm(b->scfg);
    b->norm2=b->num*factor/(unsigned)n2; /* ULong -> int ??? */
    if( (double)(b->num*factor/n2) != b->num*(double)factor/n2 ) ERR(9111);
  }
/* printf("\n dbg3: %s o=%d n2= %d norm2= %d bnum %d",ccfg(b->scfg),oo,(int)n2,(int)b->norm2,(int)b->num); */
  if(!b->norm2) b->sign=(tphase)0;
  return(b->scfg);
}

/* b = b + Tud(b), Tud is the inversion operator, b is the state 
 * called from b_getbase() for nu==nd only
 */
void b_getbase_ud(struct sbase *b){  /* for Sz!=0 wasted CPU time! */
  tbase ib0,ib1,ib2;int i,j,nj; tphase s1;
  j=nj=b->num; ib0=b->scfg; ib1=b->ib[0];
  for(i=0;i<nj;i++)/* if(b->is[i]!=0) */{
    get_sudconfig(b->ib[i],b->is[i],&ib2,&s1);	/* s|ib2>=T|ib1> */
    if(!(ib0<=ib2)){b->scfg=ib0=ib2;b->sign=s1;if(b->o==2)return; }
    if( (ib1==ib2)){
      if(!phase_eq(s1,b->is[0])){b->sign=0;b->num=0;return;}
      j=nj;break; /* because |symcfg> = Tud|symcfg> Jan00 */
    }
    b->ib[j]=ib2; b->is[j]=s1; j++;
  }
  b->norm2=b->num=j; return;
}

/* ---- long=+uudd+uudd -> short=2uudd+0uudd ----
 * faster, if symconfig and sudconfig called one after another?
 * kuerzen als extra-funktion ??? redundanter code .._lm
 * better save index for smallest? (temporarly) SORTING???
 * create cache??? or create only different configs
 * n1_sym is ca. n1/sym, d.h. ib1==ib2 | twice>1 is rare -> faster?
 *  ToDo: same as b_smallest_new()?
 * called by xval.c: modeloverlap stringorder (ToDo: remove?)
 *           spins.c: nhm_err1 storeh2 outbasis
 * in:
 * out: b->scfg=ret=mincfg; 
 *      b->ib ??? (unused? remove! ToDo)
 */
tbase b_getbase(tbase ib1,struct sbase *b){  
  tbase ib0,ib2; int j,n2,nj=sym.numall+1; tphase s1; static int err=4;
  b->ib[0]=b->scfg=ib0=ib1; b->is[0]=b->sign=1; n2=1;
  if(nj*((nu==nd)?2:1)>MaxSBase){
    E0("b-getbase: MaxSBase to small, skip it");
    if (!--err) F0("stop after 4 errors");
    return(b->scfg);
  }
  for(j=1;j<nj;){ s1=1;
    get_symconfig(ib1,((unsigned)j)-1,&ib2,&s1,1);
    if(!(ib0<=ib2)){b->scfg=ib0=ib2;b->sign=s1;if(b->o==2)return ib2;}
    if( (ib1==ib2)){
      if(!phase_eq(s1,1)){b->sign=0;b->num=0;return ib2;} n2++; }
    /* storing to ib[syms] may be memory intensive, want to remove that */
    b->ib[j]=ib2; b->is[j]=s1; /* ToDo: better call insert_sorted here */
    j++;
  }
  b->num=nj; b->norm2=nj*n2; if(n2>1) b_twice(b,n2); /* obsolete if insert_sorted */
  if (nu==nd) b_getbase_ud(b);
#if defined(CONFIG_S1SYM)    /* ToDo: does not work at the moment for tU JJ? */
  /* b_out(b,b->num); */
  if(sym.num_lm!=nn) b_getbase_lm(b,2);
  /* b_out(b,b->num); */
#endif
  return(b->scfg);
}

/* reference function for fast b_smallest3, slow, only debugging
 * ATTENTION if is[smallestcfg]=0 !!! sign of ib1
 * 1=true 0=false(smaller symcfg), -1=false(destructiv sign)
 *  ud-test only usefull if 1st spin=d!!!faster since Jan00
 */
int b_ifsmallest(tbase ib1, struct sbase *b){ 
  int old_o;
  old_o=b->o;b->o=2; b_getbase(ib1,b); b->o=old_o;
  /* printf("\ncfg=%s: ",ccfg(ib1));b_out(b,0); */
  if (!(ib1==b->scfg)) return 0;
  if (b->norm2==0 || phase_eq(b->sign,0)) return -1;
  b->sign=1; b->num=0;				/* not given here */
  return 1;
}


/*  effect of 2-site-LM-symmetry for tU-model considering fermionic sign (Feb04)
 *      .. => +..              .3 => +3.           33 => +33
 *      .u => +u.*parity       ud => -du
 *      uu => -uu vanished     u3 => +3u*parity
 *   LM = c(+,p[i])c(+,p[j])c(-,j)c(-,i)
 *  v2.21 using sym.lnext N=8 s=10/2 nud=17+10 20s(43%) => 3s(9%) PM600MHz-pg
 */
int b_ifsmallest_lm(tbase ib0){
  int i1,i2; CM({},Err17,Err17)
  for(i1=0;i1<nn-1;i1++) { i2=sym.ln[i1];
     if(i2<nn) {  /* compare member of composed spin with next member */  
      if (geti(ib0,i1)< geti(ib0,i2)) return 0;
#if CM(0,1,1) /* (+uu) + (-uu) */
      if (geti(ib0,i1)==geti(ib0,i2) && abs(zi(ib0,i1))>0) return 0;
#endif
    }
  }
  return 1;
}

/* calculate smallest lm! (enough for smallest)
   ToDo: hard code this part or examine only lm arround changed spins
      ns*(ns+1)/2 => ns, or lm*H as optimized C++ Operator
      or lm sorted Index (proof only idx-neighbours) or lm-table
 *  v2.21 using sym.lnext N=8 s=10/2 nud=17+10 7s(25%) => 4s(17%) PM600MHz-pg
 */
tbase b_smallest_lm(tbase ib0, tphase *s1){  
  int b1,b2,i1,i2;   // c=creation a=anihilation
  // table for 16 tU-possibilities (4*site1+site0)  J=t*t/U=ci.ai.cj.aj
  //   Permutation Pij = cid ciu cjd cju aju ajd aiu aid  ???? c1c2_du a2a1_ud
  //   Perm1= (getsite1+putsite1) = c1u c1d a1d a1u (ohne sign change!)
  //    JJ      __ __ __ __ __ uu ud __ __ du dd __ __ __ __ __   =  4
  //    tJ      .. .u .d __ u. uu ud __ d. du dd __ __ __ __ __   =  9
  //    tU      .. .u .d .3 u. uu ud u3 d. du dd d3 3. 3u 3d 33   = 16
  // ud-sym .... ???
  // better_fsig +        +     +  +        +  +     +        + ???
  int fsig[16]={+1,+1,+1,+1,+1,-1,-1,+1,+1,-1,-1,+1,+1,+1,+1,+1}; // sign 4+12?
  int pari[16]={ 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0}; // parity 8
  //                           ??  ?        ? ??
//  CM({},Err17,Err17)	/* signum ??? */
  if (s1) ;             /* dummy access, avoiding cc warnings */
//  printf("\n cfg_sm_lm_i= %+g %s", Re(*s1), ccfg(ib0));
  for(i1=0;i1<nn-1;i1++) { /* l1=sym.lm[i1]; */
    b1=geti(ib0,i1); if (b1<CM(2,2,3)) { /* b1={CM(u,.ud,.ud3)} ToDo */
      for(i2=sym.ln[i1];i2<nn;i2=sym.ln[i2]) /* i2>i1, if(l1==sym.lm[i2]) */ {
        b2=geti(ib0,i2);
        if (b1<b2) {
          ib0=puti(puti(ib0,i1,b2),i2,b1); b1=b2; /* exchange sym-sites */
#if CM(0,1,1) // cfg-coding independent
          if (pari[b2*4+b1]) *s1 *= parity(ib0 & p_mask[i1][i2]);
          //*s1 = fsig[b2*4+b1] * (*s1);    /* ud (fermionic exchange) */
        } else // b1 may have changed to b2! fixed 2012-07-04
        { // for uXu cfgs we expand to +uXu + -uXu which interferes to 0 ??? 
          // +uXu + -w*uXu surviving needs wLM=-1(k=pi) as LM2-Phase
          // +3X3 + +w*3X3 
    //      if (b1==b2 && abs(zi(ib0,i1))>0){
    //        *s1=0; return ib0; } /* uu + -uu */
#endif
        } /* b1<b2 or dummy */
      } /* skip max site cfgs */
    } /* i2=ln[i1]... > i1 */
  } /* i1=0..nn-2 */
//  printf("\n cfg_sm_lm_o= %+g %s", Re(*s1), ccfg(ib0));
  return(ib0);
}

/* is=0 dont store or at end, is!=0 to the top, .num < nj */

/* b->o: 0=getbase     sign,norm2,scfg,num,ib[],is[] ok
 *       1=smallest    sign,norm2,scfg ok ???
 *       2=ifsmallest  if(ret=1){ret sign,scfg,norm2} ok, num=0
 */
/* ToDo: replace by b_smallest/b_getbase + b->o=2, very similar fkt.
 * what about costs? min/mean/max clocks per call? see b_smallest() 
 * b_smallest is very similar, here we have only performance adaptions
 *  1=true 0=false(smaller permutation found) -1=false(destructive sign)
 *  -1 is returned after all ret=0 tests, else we have additional
 * return 0, not -1, for cfgs like (3.3333)
 *   this reduces compute time in new_numsymconf(v2.27)
 *   . means empty, 3 means up+down for the tU-model
 * called by ini_norm2 ns_thread
 * on S1SYM ifsmallest may fail, if other syms mix S1 subsites
 *   ToDo: Permutations must not mix sites (its bitweight) within S1 groups?
 *         do it for all commuting perms!? -> example
 *   |1234> P1=(1-3 2-4) P2=(1-2) P3=(3-4) P4=P1*P3=(1-4 2-3)=not_allowed
 * ToDo: store biggest (max) and/or smallest (min) symcfg???
 */
int b_ifsmallest3(tbase ib1, struct sbase *b){ 
  t_sites k1[NN],k2[NN],*k3,i2; int n2; register int i,j,d; tphase s1,s2,s3;
  /* move k1,k2,ks to sym.buff for L1-caching? NN*MaxSym=40*410 */
  tbase ib2=ib1;
  b->sign=1; n2=1; s2=+1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=T|ib1> */
#if defined(CONFIG_S1SYM) /* return b_ifsmallest(ib1,b); */
  if (sym.num_lm!=nn) { /* check for LM-symmetric sites */
    if (!b_ifsmallest_lm(ib1)) return 0; /* never hits!? tU? */
    if (nu==nd && sym.wud) {
       // tU 0x36ffff (LM2) generates minus sign from odd double sets
       tbase ib3=ib2;s3=s2;
       ib2=b_smallest_lm(ib2,&s2); /* smallest inverse cfg */
       if (s2==0) { /* destructive sign, tU+tJ only, never hits? */
         E1(" %s",ccfg(ib1));
         E2(" %s %+.0f",ccfg(ib3),1.*D(s3));
         E1(" %s",ccfg(ib2));
         ERR(9999); return -1;
       }
    }
  }
#endif
  if (nu==nd && sym.wud) {
    if (!(ib1<=ib2)) return 0; /* inverse config is smaller */
    if ( (ib1==ib2)) { // would also be done in the permutation loop?
      if (s2!=1) { b->sign=0; } /* destructive phase, tU only? (.333) */
      n2=2;
    }
    for(i=0;i<NN;i++) k2[i]=geti(ib2,i);
  }
#if 0 // NN%4
#error NN must be multiple of 4 (performance! unroll4) 2012-06-29
#endif
  // NN makes optimization easier for some compiler
  for(i=0;i<NN;i++) k1[i]=geti(ib1,i);  /* 30% CPU */
  for(i2=0,k3=k1,s3=1;;i2++,k3=k2,s3=s2){  /* ud=+1,-1 */
    /* search for smallest symmetric config as representator */
    for(j=0;j<sym.numall;j++){
      // sym.i[nn..NN-1] must be filled with unit operator [k]=k
      i=nn; do { i--;	/* Tj, high -> low, loop unroll2? */
        /* backward loop can be replaced by forward loop easily */
        //  4 byte = 1 int aus ram lesen und unloop 4
        /* permutation, 40%..73% CPU */
        d= (           k1[i  ])
          -( k3[ sym.i[j][i  ] ]);
      } while (d==0 && i!=0);       /* permute as long as bits are equal */
      if (d< 0) continue;           /* permutation is greater */
      if (d!=0) return 0;           /* permutation is smaller */
#if CM(1,0,0)  /* signum is not cfg-dependend */
      s1=sym.w[j];                  /* T|x>=t|x> same config */
#elif CM(0,1,1)  /* signum is cfg-dependend */
      tbase ib4; s1=1;
      get_symconfig(((i2)?ib2:ib1),j,&ib4,&s1,1); /* get sign */
#endif
      if(!phase_eq(s1*s3,1)) b->sign=0; /* return (3.3333)=0 not -1 */
      n2++;
    }
    if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 */
    IFww( break; )
  }
  b->scfg=ib1; b->num=0; /* not given here */
  b->norm2=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* OK ??? */
#ifdef CONFIG_S1SYM
  { Tnorm2 factor=b_factor_lm(ib1); /* check for 64bit overflow */
    if( (double)(b->norm2*factor) != b->norm2*(double)factor ) ERR(9112);
    b->norm2*=factor;	/* Ok ??? */
  }
#endif
  if (b->sign==0) return -1; /* Oct07 2.27 */
  /* b_out(b,0); */
  return 1;
}


/* --- last change 1998-169 JS --- */
/* static inline */ /* needed in spins.h + xval.c */
/*
   spins.h struct sbase, tbase scfg, Tnorm2 norm2, tphase sign, int num
   work variables:
    k1 - config ib1
    k2 - config ib1 or spin-inversed ib1 (not used for nu!=nd)
    ks - smallest config found actually 
   used by spins.c nhm() 107/135s, use b->sign, b->norm2 and b->scfg
   ToDo: remove b->bi[] array (for getbase call remalloc)
   see b_smallest_old (old version 10times faster)
   called by: nothing (simpler code for FPGA replacement?)
 */
tbase b_smallest_new(tbase ib1, struct sbase *b){ /* about 63% CPU -O1 10+10 */
  int j,i2,n2; tphase s1,s2,s3;
  tbase ib2, ib3, ib4;
  b->sign=1;  b->scfg=ib1; b->norm2=1;	/* default minimum */ n2=1;
  ib2=ib1; s2=1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=T|ib1> */
#if CM(0,1,1)   /* v2.0.2++ ToDo: get fast tJ8u1d2e1 running ??? */
  { b->o=1;b_getbase(ib1,b);b->o=0; return(b->scfg); } /* wont work for LM */
#endif
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){ 
                      b->scfg=    ib1=b_smallest_lm(ib1,&(b->sign));
                      if (nu==nd) ib2=b_smallest_lm(ib2,&s2);
                      if (s2==0 || b->sign==0){ b->sign=0; return TBASE0; }
                      b->norm2*=b_factor_lm(ib1); /* fermionic sign? */
                    }
#endif
  if (nu==nd && sym.wud){
    if((ib2==ib1)){ /* better do it later, because its rarely */
     n2=2;
     if(s2!=1){ b->sign=0; return TBASE0;}	/* s1!=1 */
    }
    if(!(ib1<=ib2)){ 
     ib3=ib1;b->scfg=ib1=ib2;ib2=ib3; b->sign=s2;s2=1; }	/* new minimum */
  }
  for(i2=0,s3=b->sign,ib3=ib1; ; i2++,s3=s2,ib3=ib2){  /* ud=+1,-1 */
    /* search for minimum symmetric config as representator */
    for(j=0;j<sym.numall;j++){		/* take a SymOp=PermOp(j)=ps */
      /* better call/inline get_symconfig ? */
      get_symconfig(ib3,j,&ib4,&s1,1);
      if (ib4> b->scfg) continue; /* its greater, forget it (how often?) */
      if (ib4==b->scfg) {    /* same (how often?) phase -> int */
        if(!phase_eq(s1*s3,b->sign)){b->sign=0;return TBASE0;}
        n2++; continue;
      }
      b->scfg=ib4; /* found smaller cfg */
      n2=1;
      b->sign=s1*s3; /* better use k(2PIi/l) */
    }   /* test next sym */
    if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 */
    IFww( break );
  }     /* test ud-sym */
  b->num=0;				/* not given here */
  b->norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
  /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scfg2idx(b->scfg)]; */
  return b->scfg;
}

/* called by scfg2idx_err nhm_err1 nhm hamilton_nhv? wopij_block
 *  ToDo: array of ib1's, output array, return number if calculation stopped?
 *  ToDo: better return 0 or 1=smallest to replace if_smallest
 *        input ib1 via *b (less arguments, type tbase not need to declared)
 *   10times faster than old code calling get_symconfig numsym times
 * called by: xval.c.mrule ...
 */
tbase b_smallest(tbase ib1, struct sbase *b){ /* about 63% CPU -O1 10+10 */
  /* L1-caching? NN*MaxSym=40*410 */
  int j,i2,n2; register int i/*,d*/; tphase s1,s2,s3;
  tbase ib2; register tbase ib3, ib4=TBASE0;
  /* set default minimum, better set by caller? also reduces arguments */
  b->sign=1;  b->scfg=ib1; b->norm2=1; n2=1;
  ib2=ib1; s2=1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=T|ib1> */
  /* 2014-05-24 fast hubbard (tU) does not work, fix it in 2.48 */
#if CM(0,1,1)   /* v2.0.2++ ToDo: get fast tJ8u1d2e1 running ??? */
  { b->o=1;b_getbase(ib1,b);b->o=0; return(b->scfg); } /* wont work for LM */
#endif
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){ 
                      b->scfg=    ib1= b_smallest_lm(ib1,&(b->sign));
                      if (nu==nd) ib2= b_smallest_lm(ib2,&s2);
                      if (s2==0 || b->sign==0){ b->sign=0; return TBASE0; }
                      b->norm2*=b_factor_lm(ib1); /* fermionic sign? */
                    }
#endif
  if (nu==nd && sym.wud){
    if((ib2==ib1)){ /* better do it later, because its rarely */
     n2=2;
     if(s2!=1){ b->sign=0; return TBASE0;}	/* s1!=1 */
    }
    if(!(ib1<=ib2)){ 
     ib3=ib1;b->scfg=ib1=ib2;ib2=ib3; b->sign=s2;s2=1; }	/* new minimum */
  }
  // NN faster code as nn ?
//  for(i=0;i<NN;i++){ ks[i]=k1[i]=geti(ib1,i); k2[i]=geti(ib2,i); }
  for(i2=0,s3=b->sign,ib3=ib1;;i2++,s3=s2,ib3=ib2){  /* ud=+1,-1 */
    /* search for smallest symmetric config as representator */
    for(j=0;j<sym.numall;j++){		/* take a SymOp(j)=Permutation */
      /* better call/inline get_symconfig ? */
      ib4=TBASE0;
      i=nn; do { i--;                 /* Tj, high -> low */
        /* cache miss? -> no, its accessed ca. 700 times per call, 14s */
        /* 32+8: J2=.5 0.10205747 0.25766905 53s,5.5m -> 49s,4.9m */
        /* 32+8: 700*hnz=4.4*sym*hnz max=N*sym*hnz */
        ib4 = (ib4 << CM(1,2,2))
            | ((ib3>>(sym.i[j][i]<<CM(0,1,1)))&CM(1,3,3));
        /* d= (ib1>>i)-ib4; */  /* compare cfg with SymOp*cfg */
        /* [j][i]-access vs. pointer good for x86, bad for sparc? */
        /* check if cfgs are still equal after permutation of site i */
      } while (/*d==0*/(ib1>>(i<<CM(0,1,1)))==ib4 && i!=0); /* true: 34+6=83% 32+8=78% */
      if ((ib1>>(i<<CM(0,1,1)))<(ib4)/*d< 0*/) continue; /* greater cfg, 157*hnz calls, true(32+8)=99.2% */
#if     CM(1,0,0)  /* signum is not cfg-dependend */
      s1=sym.w[ j ]; 
#elif   CM(0,1,1)  /* signum is cfg-dependend */
      get_symconfig(ib3,j,&ib4,&s1,1); /* permute sites and get sign */
#endif
      if (ib1!=ib4/*d!=0*/) {  /* 32+8 J2=.5: 1.2*hnz calls, true: 99.7% smaller cfg */
#if CM(1,0,0) /* store rest config */
        do { i--; ib4 = (ib4 << CM(1,2,2))
             | ((ib3>>(sym.i[j][i]<<CM(0,1,1)))&CM(1,3,3)); } while (i!=0);
#endif
        ib1=ib4; n2=1; b->sign=s1*s3; continue; /* found smaller cfg */
      }
      /* get same config: S|k> to s|k> */
      /* sign=0? 32+8 true=0..3% sym_k-depending, better use k(2PIi/l)? */
      if(!phase_eq(s1*s3,b->sign)){b->sign=0;return TBASE0;}
      n2++;
    }   /* test next sym */
    if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 */
    IFww( break );
  }     /* test ud-sym */
  b->scfg=ib1;
  b->num=0;				/* not given here */
  b->norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
  /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[scgfg2idx(b->scfg)]; */
  return b->scfg;
}

// next_minsymcfg: start highest (all bits 1) set 0 starting from MSB(=bitN-1)
// next_maxsymcfg: start lowest  (all bits 0) set 1 starting from LSB(=bit0) ?
//                                                       ... from MSB(bitN-1)!
// ToDo: check on 6+4 s1-chain
// next_maxsymcfg: start highest (all bits 1) set 0 starting from LSB(=bit0) 
//    stop or go deeper if found
//   
//   not easy to find, so we try to find a good upper bound
//   scan full subspace nu=4 and than walk only nu+1-max-paths
//  ToDo: get_maxscfg parallel ab nu-4 ?
//  ToDo: stepwise nu+1 statt rekursiv (parallel)
//        umbenennen in lastsymcfg ?
tbase get_maxscfg(){  /* get max(last) sym config to speedup scfg search */
  // ULong i2=0;
  int ii;
  tbase maxscfg= 0;
  ULong nt=0l, nt0=0l, nt1=0l;   /* number tested cfgs for speed analyse */
  if (vvv && !mpi_i) { printf(" get_maxscfg vvv=%d\n", vvv); }

  extern ULong x_max;
  struct sbase b;   /* working buffer (big) */
  int tnu=nu, tnd=nd; /* store the original nu,nd and restore later */
  tbase state[2*NN+1]; /* rekursion buffer (max nu+nd) */
  int id[2*NN+1], iu[2*NN+1]; /* number of bits set */
  int rd=0; /* rekursion depth */
  int rs;   /* return ifsmallest */
//  int hus[2*NN+1]; /* highest unset site(bit), goes from high to low */
  int lus[2*NN+1]; /* lowest  unset site(bit), goes from low to high ?? */
  int nbp[2*NN+1]; /* next bitpattern, 3=u+d 2=d 1=u 0=. */
  // long t2, t1=itime();
  b.o=2; /* ??? */

  /* the initial state is all sites with maximum bits set (000...000) */
  state[0]=TBASE0;  // maxscfg
  for(ii=0;ii<nn;ii++) state[0]=puti(state[0],ii,CM(2,2,3)); // 2=d minscfg
  // now we have maxcfg CM(0xff=dddddddd,0xaaaa=dddddddd,0xffff=33333333)

  // hus[0]=nn;                  // start from MSB+1=nn go down to 0
  nbp[0]=CM(0,-1,-1);         // nbp+1=CM(1=u,0=.,0=.)  goes up ++ ?
  nbp[0]=CM(2,2,3);         // nbp+1=CM(21=du,210=du.,3210=3du.)  goes down
  id[0]=nn; iu[0]=CM(0,0,nn); // ini_maxcfg
  lus[1]=lus[0]=0;            // start from LSB=0 go up to MSB=nn-1
#if CM(1,0,0)
  lus[1]=lus[0]=nu-1;  // speedup 2012-06-24, enough space for rest 0s on LSBs
#elif CM(0,0,1)
  lus[1]=lus[0]=min(nn-nu-1,nn-nd-1); // enough space for rest zeros
#endif
  
  if ((vvv&4) && !mpi_i)
    printf(" start     %s rd %2d lus %2d\n",
           ccfg(state[rd]), rd, lus[rd]);
  /* now reduce  bits from lowest bits, test if smallest and rekursiv */
  /* ToDo: hus and nn-rd is the same! remove hus[] */
  if (CM(nu==0, nu==0 && nd==nn, nu==nn && nd==nn)) {
    if (!mpi_i)printf("DBG: this trivial case of nu, nd is set extra\n");
    rs=b_ifsmallest3(state[rd],&b); nt++; /* rs: 1=true 0=false -1=destr_sign */
#ifdef CONFIG_ANSATZ
    if (!ansatz_test(state[rd])) { rs=0; }; // ToDo: untested!
#endif
    if (rs==1) maxscfg= state[rd];
  } else // trivial case
  for(;;) { /* can we do that for a thread specifique range? */
    // break if nu,nd reached!?
    // next bit pattern on current site
    //  old: JJ: [0] 1 1 ...  tJ: [-1] 0 1 0 1 ... tU: [-1] 0 1 2 0 1 2 0 1 2
    //  min: JJ: [2] 1 1 ...  tJ:  [2] 1 0 1 0 ... tU:  [3] 2 1 0 2 1 0 2 1 0
    nbp[rd]--;    /* next bitpattern, 3=u+d 2=d 1=u 0=. at site lus[rd+1] */
    if (nbp[rd]<CM(1,0,0)) {
//    lus[rd]--; /* highest unset site goes down to next lower site */
      lus[rd+1]++; /* highest unset site goes up to next higher site */
      nbp[rd]=CM(2,2,3)-1; /* next bit pattern at site lus[rd+1] */
    }
    // if (geti(state[rd],lus[rd+1])==nbp[rd]) continue; // no double set
    if (lus[rd+1]>=nn || (rd>0 && lus[rd+1]>=lus[rd]) )
    {
      /* now we have tried all sites, so we can return */
      if (rd==0) break; /* all work done */
      /* not a max, resume search at rd-1 */
      // if (rd>FullDepth) rd=FullDepth; else 
      rd--;     /* go up stack */
      if (nt<x_max && (vvv&5)==5 && !mpi_i)
        printf(" come back %s rd %2d lus %2d %2d+%2d %4.0f\n",
               ccfg(state[rd+1]), rd+1, lus[rd+1], iu[rd+1], id[rd+1],1.*nt);
      continue; /* return and try next deeper state */
    }
    /* probewise reset bit at hus/lus */
//    nbp[rd+1]=CM(0,-1,-1);
    nbp[rd+1]=CM(2,2,3)-1; // start with overflow nn to nn-1
    /* set nu, nd to have correct working behaviour of b_smallest */
    id[rd+1]=id[rd]-1+(nbp[rd]>>1); /* 1: nd--  0: nd-- */
    iu[rd+1]=iu[rd]+((1&nbp[rd])?CM(1,1,0):CM(0,0,-1));  /* 1: nu++  0: nu-- */
//    id[rd+1]=id[rd]+(nbp[rd]>>1); /* JJ: 2: nd++  */
//    iu[rd+1]=iu[rd]+((1&nbp[rd])?CM(0,1,1):CM(-1,0,0));  /* */
    if (lus[rd+1]<0) continue; // tJ, tU only
    state[rd+1]=puti(state[rd],lus[rd+1],nbp[rd]); /* lower bit pattern, down -> up */
#if CM(0,1,1)
    // dont go a wrong path 2012-07-10
    if (iu[rd+1]<tnu || id[rd+1]<tnd) continue;
#endif

#if 1
    // 39+1 speedup, 37+17 speedup 21s 3880724 of 10988161 (old 37s)
    // if there is only one choice, skip all pre-cfgs and go to final cfg
    CM(\
    if (lus[rd+1]>0 && tnu-iu[rd+1]==lus[rd+1]){rd++;continue;},{},{})
/*
    CM(\
    if (lus[rd+1]>0 && id[rd+1]-tnd==lus[rd+1] \
                   ) { rd++; continue; },{},{})
*/
#endif

    // prepare for going deeper
    lus[rd+2]=0; /* start from lowest site when going deeper */
#if CM(1,0,0)  // 2012-06-24 speedup for nu>>nd 39+15 13s, 19+35 1s
    // 14+2: 0x7fff 14-1-3 ??? 
    lus[rd+2]=tnu-iu[rd+1]-2; // (much) faster if starting on higher bits
#elif CM(0,0,1)
    lus[rd+2]=min(iu[rd+1]-tnu,id[rd+1]-tnd)-1; // -3 ??? 0..nn
    // printf(" rd+2= %d  lus= %d\n",rd+2,lus[rd+2]);
#endif
    /* ToDo: filter thread specifique range here? */
    nu=0; nd=nn; /* trick to use sym_ud to early */
    if (iu[rd+1]==tnu && id[rd+1]==tnd) {nu=tnu; nd=tnd;} /* sym.ud=on */
    rs=b_ifsmallest3(state[rd+1],&b); /* 1=true 0=false -1=destr_sign */
    nt++; /* count ifsmallest calls for speed output */
    if (((rs && iu[rd+1]==tnu && id[rd+1]==tnd && (nt0<x_max || !(nt0&(nt0-1))))
       || nt<x_max || !(nt&(nt-1))) && (vvv&5)==5 && !mpi_i) // || hus[rd+1]!=nn-rd-1)
      printf(" mtry % d   %s rd %2d lus %2d %2d+%2d %4.0f\n",
        rs, ccfg(state[rd+1]),rd+1, lus[rd+1],iu[rd+1],id[rd+1], 1.*nt);
    if (rs) { /* go deeper or count for 1=true and -1=destructiv_sign */
      if (iu[rd+1]==tnu && id[rd+1]==tnd) { /* found symmmetric state */
#ifdef CONFIG_ANSATZ
        if (!ansatz_test(state[rd+1])) { rs=0; } // untested
#endif
        nt0++; // count tests for reduced debug output
        if (rs==1) {  /* without destructive phase */
          // tbase old= maxscfg;
          if (maxscfg < state[rd+1]) maxscfg= state[rd+1];
          if (nt>1e9) E0("maxscfg stopped after 1e9 tests");
          if (rd==0 || nt>1e9) break;
          break; // 2012-06-24 fixed version
          // break; /* ToDo: return to next subtree tnu=4 */
        } /* rs==1 found smallest, nondestructive sign */
      } else { /* found symmetric mother cfg, go deeper */
        nt1++;
        rd++; /* go deeper, ToDo: start by putting rest up's to next sites? */
        if (rd>2*NN) { ERR(9999); } continue;
      }
    } /* smallest cfg */
  } /* for(;;) */
  nu=tnu; nd=tnd; /* restore gobal nu, nd */

  if (!mpi_i) {
    printf("  end.maxscfg nt0= %lu nt1= %lu nt= %lu t+= %s\n",
      nt0, nt1, nt, sclock(~0l));
    printf("  maxscfg= %s\n", ccfg(maxscfg));
  }
  if (maxscfg==TBASE0 && !mpi_i) E0("bad maxscfg code?, ignore this?"); 
  return(maxscfg);
}


#define BLOCKS	8	/* can be only 2^n=1,2,4,8,etc */
/* version 0.1 Feb 2000 first try version, not finaly, for checks 
 * ideal way for part l1 into blocks => multi processing
 * in ideal case fast and invariant against symmetry operations 
 * min=0 max=n*(n-1)/2 ??? besser mit Untergittern
 */
int block_no(tbase ib1){ /* use pair correlations, count d-d pairs */
  int no,i1,i2; no=0;
  for(i1=0;i1<nn;i1++) if (zi(ib1,i1)==-1) /* nn+nd*nn Ops */
  for(i2=0;i2<nn;i2++) if (zi(ib1,i2)==+1)
   no+=sym.corr[i2][i1];
  return (no>>1)&(BLOCKS-1);
} 

/* --- all what be needed for pthreads of numsonf --- */
typedef struct sp_ns {
#ifdef CONFIG_PTHREAD
  pthread_t pt;
#endif
  int block;       /* block nr (0..threads*nodes-1) for debugging */
  ULong i1,n;      /* idx range of configuration to check (rw,r) [i1...n-1] */
                   /* ToDo: on return i1 is i1(nothing done)...n(all done) */
                   /* ToDo: return time too, for dynamic balance */
                   /* ToDo: timeout in seconds dflt=60 */ 
  ULong numans;    /* num_ansatz (w) symcfgs filtered out by ansatz_test() */
  Tnorm2 norm2;     /* check sum (w) */
  ULong num;       /* number of stored sym configs (w) */
  ULong num0;      /* number of sym configs with sign=0 (w) v2.27 */
  tbase ib1;        /* spin-configuration (t) */
  struct sbase b;   /* working buffer (t,big) */
  struct sp_ns *next; /* pointer to next data set */
  // char fname[32];     /* filename tmp_shared/l1_%d%4d.tmp (r) Oct07,Jan12 */
  unsigned char *buf; /* space for results (sym-configs) (r[w]) 2012-04 */
  ULong buflen;       /* size of buffer in units of Bsize (r) 2012-04 */
  int time0;          /* computing time in seconds to calculate balance */
} tp_ns;

/* void-function for pthreads called by numsymconf(), see below */
#define L1_BUFLEN (8*1024*1024)  /* in units of Bsize, ca. 40MB for 1-10write/s */
void ns_thread( void *ptr ){
    tp_ns *p=(tp_ns *)ptr;
    tbase ib0;      /* min config (overflowtest) */
    tbase ib_stop;    /* in case of s=1 next skips LM-cfgs 2012-01 */
    ULong i1 /*, i2, i3*/;

    long t2,t1=itime();  // ToDo: syscalls only if ptr->tflag=1 
    ULong i_out=1l<<14; /* counter for output, speed estimation */
    int rt;              /* return code from b_ifsmallest3 */
    /* 2012-01 p->n may be to big, in case LM-cfgs are skipped by next() */
    /* ll1(n) may cause trouble on s=1 and N=64 bit systems */
    ib_stop = ll1(p->n /* - 1*/); /* overrun nosym_n1? 2012-03: ll1(n1)==0 bad!? */
//    ib_stop = next(ib_stop); /* trick to handle overflow correct */
    p->ib1 = ll1(i1= p->i1); /* start cfg */
//  if (i1) { p->ib1 = next( ll1(i1-1) ); } /* trick to get correct s1sym */
// ToDo: 9s1=3s3 next(aaaaaaaaabbbbbbbbb
//                   =aaabbbcccdddeeefff
//                   =100000000011111111) may fail (noS1sym)! N=2*9 2012-05-16
    ib0 = ll1(0);
    if (!ib_stop) ib_stop=ib0;
    /* but its better to do this outside run, because s=1 overlap problem */
#if 0
    // if (mpi_i<=8) 
    {
      usleep(mpi_i*10000);
      printf("DBG mpi %d %s - %s i %9lu - %9lu\n",
        mpi_i, ccfg(p->ib1), ccfg(ib_stop),
               (unsigned long)(p->i1), (unsigned long)(p->n));
    }
#endif
    if    (i1 < p->n) // JS20120306 minimum condition
    for( ; i1 < p->n ; ){
      rt=b_ifsmallest3(p->ib1,&(p->b)); /* 1=true, 0=false -1=sign0 */
#ifdef CONFIG_ANSATZ
      if (!ansatz_test(p->ib1)) { p->numans++; rt=0; }
#endif
//printf("DBG i1= %6.0f %s = %d\n", 1.*i1, ccfg(p->ib1), rt);
      if (1==rt) { /* 1=true, 0,-1=false */
        if (p->num >= p->buflen) break;
        putl1(/*i2*/p->num,p->ib1,p->buf);   /* store mincfg to l1p[l1s] */
        p->num++; /* i2++; */            /* count stored cfgs */
        p->norm2+=p->b.norm2;
      }
      if (-1==rt) { p->num0++; p->norm2+=p->b.norm2; }
      i1++;                     /* loop-counter */
      p->ib1=next(p->ib1);      /* next higher config (skip LM-cfgs!?) */
//printf("DBG i1= %6.0f %s stop=%s ib0=%s\n",
//   1.*i1, ccfg(p->ib1), ccfg(ib_stop), ccfg(ib0));
      /* for 2s=2,3,4... p->n is bigger than the number of next calls */
      /* next() skips LM configs, so the of= value and 100% estim. fails */
      /*  may be next works to &(p->ib1) and returns skipped numbers
       *    also getting index difference between ib1 and ib2 would be nice 
       * next may jump over ib_stop for s=1 (usually s=1/2 cant jump)
       */
      if (p->ib1 <= ib0) break;     /* avoid overrun? */
      if (p->ib1 >= ib_stop
       && ib_stop != ib0) break; /* part of next thread (s=1,3/2...) */ 
      if (i1==i_out && mpi_i==0 && p->block==0) { long t2=itime();
        i_out<<=2; /* x4 */
        if (t2-t1>8)
        /* i1 is incorrect for LM, ll2 can be used for correct estimations */ 
        printf(" ns %6.2f%% found= %9.0f of %9.0f t= %5.2fm t100%%= %6.2fm\n",
           (i1 - p->i1)*100./(p->n - p->i1),  // % of work?
             1.*p->num,                     // found sl1
             1.*i1, /* i1 < p->n - p->i1 for s=1++ */
              (t2-t1)/60.,
           // (t2-t1)/60.*p->n/i1);
              (t2-t1)/60.*(p->n - p->i1)/(i1 - p->i1)); // JS20120306
      }
    }
    /* this was buggy before 2012-01 and therefore disabled for s=1 pt>1 */
    /* ToDo: output only min/mean/max + time in the main thread later */
    /* MPI: "\n ... " may buffer ... until end (SC), better " ... \n" */
    t2=itime();
    p->time0 = t2-t1;
#if 0
#if 0
    if ((mpi_i*pt_n+p->block%pt_n<2 /* 2012-05 output first and last two */
      || mpi_i*pt_n+p->block%pt_n>mpi_n*pt_n-3)
        && (p->num>0 || t2-t1>10)) /* but not the less interesting things */
#endif
      printf(" ns_t %2d %2d found= %9.4g of %9.4g of %9.4g t=%s\n",
        mpi_i, p->block%pt_n, 1.*p->num, 1.*i1, 1.*(p->n - p->i1),
        // p->i1*1., 1.*p->n, /* for debugging s=1 from-to */
        sclock(t2-t1)); /* block=0..mpi_n*B_NUM-1 */
#endif
    /* i3=fwrite(buf,Bsize,i2,fl1); if(i3!=i2)F1("fwrite %s",p->fname);
     fclose(fl1); */ /* flush */
    /* free(buf); 2012-04 */
}  // ns_thread

/* number of (symmetric) base-states(NN,nu,nd) -> n1_using_Sym 
 *  global vars: sym,nu,nd!
 * until v2.19:
 *   dividing in equal blocks for threads is unbalanced (2/3 only one cpu)
 *   because 1st thread has more work as the others (3 times more)
 * ToDo: better thread-balance
 *   1) create nt=4 threads, set results=0
 *      pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
 *      better pthread_mutex_init(&mutex1,NULL)
 *   2) pthread_mutex_lock (only one thread can pass)  (OMP_critical)
 *   3) store results, reset results, get new work=block(ns/(16*nt)) (serialized)
 *   4) pthread_mutex_unlock
 *   5) does the job and return to (2) -or- leave if work=0
 *   ... (MPI_Barrier, OMP_barrier)
 *   6) destroy threads
 * exsample: jj square 160sym nn=36+4 sym_k=0 ns(91390)=632=523+85+22+1+1
 * ToDo: try IsingGS + non-sym-excitations until IsingMax (how to divide?)
 * ToDo: dont use disk (MPI+NFS is slow and instable)
 *    better use MPI-disk emulation OR parallel CPU and save local part
 *    store buffered blocks to node (i++%mpi_n)?
 * ToDo: transfer to local disks of cluster nodes
 *       get final_n1=sum(local_n1)
 *       + send local_n1 (0 to n0, ... local_n1 to n3, 0 to rest
 *       receive from n0,n1,n2 ... until final_n1/tasks reached
 * ToDo: estimation of n1? upper boundary?
 *        MC? 100*blocks random symmetrized startvec
 *            divided into blocks + storeh2?
 *            hash = ising+meanerg_of_excitations
 *  (minimize off_diag_block_elements for maximum performance)
 *   
 */
// ToDo: new pipelined recursion
// l1_0(u=0) -> task0(u=1) -> task1(u=2) -> task2+3(u=3) -> task4-7(u=4) -> l1_4(u=4)
//   repeat recursion using l1_4 (einfacherer [paralleler] code)
//  ToDo: chpt, read last l1_0000.dat + test if nu+nd+symcfg = OK
//     + continue from last symcfg+1
ULong numsymconf(){  /* short: ns() */
  ULong i2=0, i3, numans=0, num0=0;
  Tnorm2 sum=0;
  tp_ns pns, *ppns, *p2;
  tbase maxscfg0,   // ToDo: upper boundary (is: maybe max)
        startcfg0,  // startcfg, resumed after checkpointing, default=ll1(0);
        blast;      // checkpointing, last written valid scfg
  int ii, i4, nsalgo=0;
  FILE *fl1=NULL;
  long l1=itime();
extern int algo;
  p2=NULL ; /* dummy access, avoiding cc warnings */
  ppns=&pns; ppns->buflen=0; ppns->buf=NULL;
  i3=numconf(); i4=pt_n*mpi_n; /* i3=(N!)(nu!nd!) i4=num_threads */

  maxscfg0= get_maxscfg(); /* 2012-05-24 for scan cfg speedup */

  if (checkpoint>=2) { //  try reading n1 from chkpt2
    FILE *f1; int rt=0; n1=0;
    if (!mpi_i) { if (vvv) printf(" chkpt2.r");
      f1=fopen(FCHKPT"2","r"); if(!f1){E0("chkpt2.o");rt=1;}
      if (f1 && fscanf(f1,"%lu",&n1)!=1) {E0("chkpt2.r");rt=2;n1=0;} // 2012-05
      if (f1) fclose(f1);
    } // ToDo: test last symconfig!?
    MPI_Bcast(&n1,1,MPI_UNSIGNED_LONG,0,MPI_COMM_WORLD);
    if (n1==0) checkpoint=1; // failed, restart calculation
    // ToDo check correct l1.dat (below?)
  }
  // 2012-06-06 read last state
  blast=0;
  if(!mpi_i){ // check point resume? read l1.end and progress
    tbase b0=TBASE0;  // must be initialized with 0, see fread(..,Bsize,...) ??? 2012-06
    off_t fl2o; int fl2, rt=0,i,td; struct sbase b; // long t1;
    b.o=2; /* ? */
    fl2=open("tmp_shared/l1_0000.dat",O_RDWR); // seq. written 
    if (fl2!=-1) {
      // EOF-Bsize may fail if parts of Bsize=f(NN) has changed! 2012-06-19 len%8=6 %6=4
// * fseek(long)  32bit or 64bit (v2.32 and less)
// * lseek(off_t) 32bit or 64bit (also 64bit for 32bit + LFS) v2.33
      fl2o=lseek(fl2,(off_t)0,SEEK_END);
      if (fl2o==-1){ E0("lseek");
      } else { //ok 
        if (fl2o%Bsize) {
          printf(" WARNING: l1-file size %% Bsize%d = %d"
                 " trying to cut ...\n", Bsize, (int)(fl2o%Bsize));
          if (ftruncate(fl2,fl2o-(fl2o%Bsize))!=0) E0("truncate");
          fl2o-=fl2o%Bsize;
        }
        fl2o=lseek(fl2, (off_t)(-Bsize), SEEK_END)+Bsize;
        //  ToDo: cut to ftell%BSize==0 2012-06  (writing multiple of Bsize=7 + chkpt1)
        if (read(fl2,&b0,Bsize)==Bsize) {  // may read only Bsize=7 of 8 bytes ???
          for (td=i=0; i<nn; i++) td += 1&(b0>>(CM(i,2*i+1,2*i+1)));
          rt=b_ifsmallest3(b0,&b); /* 1=true, 0=false -1=sign0 */
          printf(" chkpt12.r_last %s td=%d scfg?=%d n1?=%ld %%%d=%d\n",
           ccfg(b0), td, rt, (long int)(fl2o/Bsize), Bsize, (int)(fl2o%Bsize));
          // ToDo: check if valid scfg rt==1 td==nd,
          //       then start search from it
          if (td!=nd || rt!=1) {
            if (checkpoint>=1) {
              E0("bad chkpt1 (rename), NN+Bsize changed?"); 
              rename(FCHKPT"1",FCHKPT"1.old"); // next start may succeed
              // MPI_Abort(MPI_COMM_WORLD,1);
            }
            n1=0; checkpoint=0; // ToDo: restart?
          }
          if (checkpoint>=2 && fl2o*1. != n1*1.*Bsize) {
            n1=0 /* fl2o/Bsize */; checkpoint=1;
            printf(" WARNING: bad chkpt2 (rename), NN+Bsize changed? fsize?\n");
            rename(FCHKPT"2",FCHKPT"2.old"); // next start may succeed
            // MPI_Abort(MPI_COMM_WORLD,1); // ToDo: restart?
          } 
          // ToDo: try to load chkpt1a (letzter gepruefte config) hex
          if (checkpoint>=1) blast= b0; // valid
        } // fread ok
      } // fseek ok
      close(fl2); fl2=0;
    }
  }
  MPI_Bcast(&checkpoint,1,MPI_INT,0,MPI_COMM_WORLD);
  if (checkpoint>=2) return (n1); // skip computation
  
  startcfg0= ll1(0);   // 2012-06-11 default start configuration (no syms)
  if (checkpoint>=1) {
    FILE *f1; tbase t0=0; int i1,rt=0,tu,td; char s1[NN+2]; // + LF + \0
    if (vvv && !mpi_i) fputs(" r."FCHKPT"1\n",stdout); // meaning changed (old n1)
    n1= 0;
    if (!mpi_i) {
      f1=fopen(FCHKPT"1","r");
      if (!f1) { printf("\nERROR o."FCHKPT"1\n"); rt=1; }
      // read ooooxoxoxoxxx or ...ooxxQQoxQ + \0
      // ToDo: additional read numstoredcfgs
      if (f1 && !fgets(s1,nn+1,f1)) {E0("r."FCHKPT"1");rt=2;} // 2012-06
      if (rt==0) {
        if (s1[0]=='0' && s1[1]=='x') { /* new hex format 2012-06-27 */
          for (i1=0;i1<(nn+3)/4;i1++)
            { t0<<=4; t0|=s1[2+i1]-((s1[2+i1]<'a')?'0':'a'+10); }
          for (tu=td=0,i1=0;i1<nn;i1++) {
            if (geti(t0,i1)&1) tu++;
            if (geti(t0,i1)&2) td++;
          }
        } else /* old site wise format */
        for (tu=td=0,i1=0;i1<nn;i1++) {
          t0<<=CM(1,2,2);
          if (s1[i1]=='.') {t0|=CM(-1,0,0);          continue; }
          if (s1[i1]=='o') {t0|=CM( 0,1,1);tu++;     continue; }
          if (s1[i1]=='x') {t0|=CM( 1,2,2);     td++;continue; }
          if (s1[i1]=='Q') {t0|=CM(-1,3,3);tu++;td++;continue; }
          E0("chkpt1: bad format (ignore chkpt)"); rt=3; break;
        }
        if ((tu!=nu || td!=nd)) { E0("chkpt1: bad u+d (ignore chkpt)"); rt=4; }
        if ((tu==nu && td==nd) && i1==nn) { 
          startcfg0= t0; // valid startcfg
          printf(" ckpt1cfg= %s dflt_start\n", ccfg(startcfg0));
          printf(" l1endcfg= %s higher?\n", ccfg(blast));
        }
      } // rt==0 no errors
      if (f1) fclose(f1);
      if (rt==0 && blast>startcfg0) {
        startcfg0=next(blast);
        printf(" startcfg= %s (l1end+1)\n", ccfg(startcfg0));
      } // else  checkpoint=0; // needed?, not if chkpt2 + l1 is ok
    } // !mpi_i
  } // chkpt1
  
  MPI_Bcast(&startcfg0,Bsize,MPI_BYTE,0,MPI_COMM_WORLD);
  MPI_Bcast(&checkpoint,1,MPI_INT,0,MPI_COMM_WORLD);
  if (!mpi_i)
    printf(" startcfg= %s chkpt=%d\n", ccfg(startcfg0), checkpoint);
  /* the estimation of n1 failes if there are local symmetries. n1 bigger? */
  /* we now try to choose the faster algorithm automatically JS 2011-10-28 */
  /* nsalgo: 0 = conventional check (parallel), 1 = tree (serial only) */
  /* for single threads switch to ca. 10*faster serial code JS 2011-10-28 */
  if (mpi_n==1 && pt_n==1) nsalgo = 1;  /* serial-tree-like-scfg-search */
  /* but user can switch to the other mode by algo+=16 */
  if (algo&16) nsalgo ^= 1; /* explicite switch to other algo for tests */
  if (vvv && !mpi_i) {
    printf(" --- starting hilbert space generation ---\n"
           " filter out symmetric ising base states (n1 .ge. N!/(nu!nd!))/sym\n");
    if (algo&16) printf("WARNING: a16 switches to the opposit of the auto"
       "-choosen mode (for testing only)\n");
    printf(" using %s algorithm\n",((nsalgo)?"serial-recursive "
                                            :"parallel-scanning"));
    /* JS-Apr2011: mpi*pt=i4 removed (diff) */
    printf(" numsymconf(%.15g)/(mpi*pt) out=tmp_shared/l1_0000.dat\n", i3*1.);
    printf(" ncfgs=%.2e nsym=%d estimated nscfgs n1: %.2e\n",  
             i3*1., sym.numall+1, i3*1./(sym.numall+1));
    /* example: chain6=3+3 n1: 6!/(3!3!)/sym=20/(6..12)=0..3 */
  }

  MPI_Barrier(MPI_COMM_WORLD);

  if(nsalgo==1 && !mpi_i)  /* new serial recursive method v2.27 */
  { /* ToDo: tJ + tU, nd>nu is much faster, test=algo&16
     * malloc 1M per thread for work, 2nd 1M block chain for storage and
     *     do parallel 0,1M,2M,...(n-1)M(1M),(concat to node i until full)
     *                 (n+0)M,(n+1)M,(n+2)M... ...
     *     balance to n1/threads or Hnzx/thread
     * number of b_ifsmallest calls dominates compute time
     *        so we have a speedup for sym>=2
     * time: old=n1/600000[min] new=sum[i=0..nu] n1s(i)/96e6
     * (40!/8!/32!)=76904685=480654*160+45=482184+482139+480948+...
     * JJ worst case: numsym=0 -> 1..10% slower
     * sym_k != 0 works, ToDo: tU,tJ
     * ToDo: if nu>nd do inversion! tU -> .3 inversion + U -> -U
     *                                for U*(niu-1/2)*(nid-1/2)==Ham2
     *        for nu>nd do inversion before nsymconf and again before storing
     *
     *  see ../doc/numsymcfg.html (explanation+speed+test+examples)
     *   tJ code checked, tU checked (tune+further checks later!)
     *
     * ToDo: simple but faster parallel partitioning (2012-01)
     *       moeglichst Reihenfolge erhaltend
     * (udddd) cpu0-7
     * (uuddd) cpu0-6 + ududd..udddu (cpu7=3+2+1)
     * (uuudd) cpu0-5 + uudud..uuddu (cpu6=2+1)
     * (uuuud) cpu0-4 + uuudu (cpu5)
     * ...   
     * ToDo2012-05: parallel-rekursiv: (worst case = 0sym, best case = s=4)?
     *   - step1: 1cfg 1+39 to file0
     *   - step2: read file0, compute 2+38 parallel to file1
     *   - step3: read file1(2+38) file1, split +compute 3+37 parallel to file0
     *   - etc. (per ping-pong-piping+sorting 2 task-groups of n/2?)
     *  or
     *   task0 1+39 pipes to task1 which does 2+38 pipe to task2 ... (speed2?)
     *   same with splitting? 2+38 to task2+task(2+(n-2)/2)=tree
     *   n-files which must be concat
     * ToDo: checkpointing!? + recompute on bad checkpoints
     */
    extern ULong x_max;
    struct sbase b;   /* working buffer (big) */
    int tnu=nu, tnd=nd; /* store the original nu,nd and restore later */
    tbase state[2*NN+1]; /* rekursion buffer (max nu+nd) */
    int id[2*NN+1], iu[2*NN+1]; /* number of bits set */
    int rd=0; /* rekursion depth */
    int rs;   /* return ifsmallest */
    int hus[2*NN+1]; /* highest unset site(bit), goes from high to low */
//  int lus[2*NN+1]; /* lowest  unset site(bit), goes from low to high */
    int nbp[2*NN+1]; /* next bitpattern, 3=u+d 2=d 1=u 0=. */
    unsigned char *buf; ULong bi2; /* write buffer and its counter */
    tp_ns *p=&pns;
    long  t1=itime();
    // ULong i_out=1L<<8; /* output of estimated time */
    ULong nt=0l;        /* number tested cfgs for speed analyse */
    if (!mpi_i) printf(" running fast recursive ns(), but single_thread\n");
    if (vvv && !mpi_i)
      printf(" new_numsymconf(%.15g)/(mpi*pt) nd>nu,"
       " nQ>n. is (1-8x) faster\n", 1.*i3); /* JS-Apr11 i4 removed (diff) */
    p->norm2=0;   p->numans=0; p->b.o=2; p->num=0;
    // sprintf(p->fname,"tmp_shared/l1_%04d.dat",0);
    // 2012-06-11 append if checkpoint>=1
    checkpoint=0; // ToDo
    fl1=fopen("tmp_shared/l1_0000.dat",((checkpoint)?"ab":"wb"));
    if (!fl1) F0("fopen_l1"); bi2=0;
    // can happen if Bsize=7 write_buffer=512 + ^C + chkpt.resume 2012-06-20
    if (ftell(fl1)%Bsize) fseek(fl1, 0L-Bsize-ftell(fl1)%Bsize, SEEK_END);
    buf=(unsigned char *)malloc(L1_BUFLEN*Bsize); if (!buf) F1("malloc node %d",mpi_i);
    b.o=2; /* ??? */

    /* the initial state is all sites with maximum bits set (111...111) */
    state[0]=TBASE0;
    for(ii=0;ii<nn;ii++) state[0]=puti(state[0],ii,CM(2,2,3)); // JJ: 2=d
    // now we have maxcfg CM(0xff=dddddddd,0xaaaa=dddddddd,0xffff=33333333)
    
    hus[0]=nn;                  // start from MSB+1=nn go down to 0
    nbp[0]=CM(0,-1,-1);         // nbp+1=CM(1=u,0=.,0=.)  goes up ++ (min-1)
    hus[0]=nn+1;                // start from MSB+1=nn go down to 0
    nbp[0]=CM(1,1,2);         // nbp+1=CM(1=u,0=.,0=.)  goes up ++ (max-1)
    id[0]=nn; iu[0]=CM(0,0,nn); // ini_maxcfg
    
    // ToDo: prepare hus,nbp,id,iu,state[] dependend from startcfg0
    
    i2=0; num0=0; sum=0; /* counter scfg states + scfgs_with_sign0 */
    if (vvv&4)
      printf(" start     %s rd %2d hus %2d\n", ccfg(state[rd]), rd, hus[rd]);
    /* now reduce bits from highest bits, test if smallest and rekursiv */
    /* ToDo: hus and nn-rd is the same! remove hus[] */
#if 1
    /* nu,nd=0,40 will not work else Apr09 */
    if (CM(nu==0, nu==0 && nd==nn, nu==nn && nd==nn)) { // n1=1 only maxcfg
      printf("DBG: do trivial case endcfg=startcfg extra\n");
      rs=b_ifsmallest3(state[rd],&b); nt++; /* 1=true 0=false -1=destr_sign */
#ifdef CONFIG_ANSATZ
      if (ansatz_test(state[rd])) rs=0; // ToDo: untested!
#endif
      /* nondestructive phase, store mincfg=state[rd] to l1[bi2*BSize] */
      if (rs== 1) {putl1(bi2,state[rd],buf);i2++;p->num++;bi2++;p->norm2+=b.norm2;}
      if (rs==-1) { num0++; p->norm2+=b.norm2; } /* count sign0 */
    } else
#endif
    // ToDo: check first for valid/go_deeper and do next/break after that
    //  from dddd to dddd=break
    //  from uddd via dudd ddud dddu to uddd=break
    //  from uudd via uuud uudu udud uduu uddu uddd=break
    for(;;) { /* can we do that for a thread specifique range? */
      // next bit pattern on current site
      //   JJ:  [0] 1 1 ...   tJ: [-1] 0 1 0 1 ... tU: [-1] 0 1 2 0 1 2 0 1 2
      // ToDo: check tU, should it not go down ([3] 2 1 0 2 1 0 ...)
      nbp[rd]++; // JJ: 1=u 2=d 3to1=u 2=d ... 
      if (CM(tnu!=0,tnd!=nn,(tnu!=nn || tnd!=nn)) // exception n1==1
       && nbp[rd]>CM(1,1,2)) {
        hus[rd]--; /* highest unset site goes down to next lower site */
        nbp[rd]=CM(1,0,0); /* next bit pattern JJ: u */
      }
      if (hus[rd]<=0 CM(|| tnu-iu[rd]>hus[rd],\
                        || id[rd]-hus[rd]>tnd || id[rd]<tnd || iu[rd]>tnu || iu[rd]+hus[rd]<tnu || id[rd]+iu[rd]<tnu+tnd,\
                        || hus[rd]<iu[rd]-tnu || hus[rd]<id[rd]-tnd)) {
        /* now we have tried all sites, so we can return */
        if (rd==0) break; /* all work done */
        rd--;     /* go up */
//        if (i2<x_max && vvv&4)
//          printf("\n come back %s rd %2d hus %2d %2d+%2d %.0f",
//                 ccfg(state[rd]), rd, hus[rd], iu[rd], id[rd],1.*nt);
        continue; /* return and try next deeper state */
      }
      /* probewise reset bit at hus */
      hus[rd+1]=hus[rd]-1;/* go to next lower site index when going deeper */
      nbp[rd+1]=CM(0,-1,-1); // min-1=max
//      hus[rd+1]=hus[rd];/* go to next lower site index when going deeper */
//      nbp[rd+1]=CM(1,1,2); // max-1
      /* set nu, nd to have correct working behaviour of b_smallest */
      id[rd+1]=id[rd]-1 +(nbp[rd]>>1); /* npb=0,1: nd-- */
      iu[rd+1]=iu[rd]+((1&nbp[rd])?CM(1,1,0):CM(0,0,-1));  /* 1: nu++  0: nu-- */

      // 2012-05 add {} for empty macro
      CM({},if (iu[rd+1]!=tnu && id[rd+1]==tnd) { continue; },{})
      CM({},{},if (iu[rd+1]<tnu || id[rd+1]<tnd) { continue; })
      /* much faster for N=40 1u1d */
      CM({},{},if (iu[rd+1]-hus[rd+1]>tnu || id[rd+1]-hus[rd+1]>tnd) { continue; })
      CM({},if(id[rd+1]<tnd || iu[rd+1]>tnu || iu[rd+1]+hus[rd+1]<tnu || id[rd+1]+iu[rd+1]<tnu+tnd) continue;,{})


      state[rd+1]=puti(state[rd],hus[rd+1],nbp[rd]); /* lower bit pattern, down -> up */
      /* go deeper if we have only one possibility to set the rest cfg */
      CM({},{},\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      && iu[rd+1]-tnu==hus[rd+1]) { rd++; continue; })
      CM({},{},\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      && iu[rd+1]-tnu==0) { rd++; continue; })
      CM({},{},\
      if (hus[rd+1]>0 && iu[rd+1]-tnu==hus[rd+1] \
                      && id[rd+1]-tnd==0) { rd++; continue; })
      CM({},{},\
      if (hus[rd+1]>0 && iu[rd+1]-tnu==hus[rd+1] \
                      && id[rd+1]    ==hus[rd+1]) { rd++; continue; })
#if 0
      CM({},{},\
      if (hus[rd+1]>0 && iu[rd+1]    ==hus[rd+1] \
                      && id[rd+1]-tnd==hus[rd+1]) { rd++; continue; })
#endif
      // 38+1
      CM({},\
      if (hus[rd+1]>0 && id[rd+1]-hus[rd+1]==tnd \
                      ) { rd++; continue; },{})
#if 0
      // 38+1
      CM({},\
      if (hus[rd+1]>tnd && id[rd+1]==hus[rd+1] \
                      ) { rd++; continue; },{})
#endif

#if 1  // 2012-05-25 no speedup for nu<nd, speedup 2..3 for nu>nd (10x slower)
      // 39+1 speedup
      CM(\
      if (hus[rd+1]>0 && tnu-iu[rd+1]==hus[rd+1] \
                      ) { rd++; continue; },{},{})
#endif
#if 1 // useless? s1-triangle 46+8 2.3m vs. 0.28m 2012-07-03
      CM(\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      ) { rd++; continue; },{},{})
#endif

      /* ToDo: filter thread specifique range here? */
      nu=0; nd=nn; /* trick to use sym_ud to early */
      if (iu[rd+1]==tnu && id[rd+1]==tnd) {nu=tnu; nd=tnd;} /* sym.ud=on */
      rs= b_ifsmallest3(state[rd+1],&b); /* 1=true 0=false -1=destr_sign */
      // rs|=b_ifsmallest3(state[rd+1],&b); /* 120629 sym=324 4.7/2.38m=1.97 */
      nt++; /* count ifsmallest calls for speed output */
      if ((i2<x_max || !(nt&(nt-1))) && (vvv&5)==5) // || hus[rd+1]!=nn-rd-1)
        printf(" try  % d   %s rd %2d hus %2d %2d+%2d %4.0f %4.0f t/m=%6.2f\n",
          rs, ccfg(state[rd+1]),rd+1, hus[rd+1],iu[rd+1],id[rd+1], 1.*nt,
          1.*i2+((rs==1 && rd+1>=tnu)?1:0), (itime()-t1)/60.); // found
      if (rs) { /* go deeper or count for 1=true and -1=destructiv_sign */
        if (iu[rd+1]==tnu && id[rd+1]==tnd) { /* found symmmetric state */
#ifdef CONFIG_ANSATZ
          if (!ansatz_test(p->ib1)) { p->numans++; rs=0; }
#endif
          if (rs==1) {  /* without destructive phase */
            if (bi2 >= L1_BUFLEN) {  /* write out buffer, disk full? */
              if (fwrite(buf,Bsize,bi2,fl1)!=bi2) F0("fwrite l1");
              bi2=0;
            }
            putl1(bi2,state[rd+1],buf);   /* store mincfg to l1p[l1s] */
            i2++; p->num++; bi2++; p->norm2+=b.norm2; /* count stored cfgs */
            /* found: hus[rd+1]=nn-(rd+2) */
          } /* rs==1 found smallest, nondestructive sign */
          if (rs==-1) { num0++; p->norm2+=b.norm2; } /* count sign0 */
        } else { // dive into the deeper subtree
          /* found symmetric mother cfg, go deeper */
          rd++; /* go deeper, ToDo: start by putting rest up's to next sites? */
// v2.38 gave n1=0 where it should not, nut sure if I fixed this correct
//          if (rd>nu && CM(1,0,0)) break; // fix nud=0,40 endless 2^40 loop 
//          if (rd>tnu && CM(1,0,0)) break; // fix nud=0,40 endless 2^40 loop 
            /* ToDo: store found+nt, on return print found as success rate */
          if (rd>2*NN) { ERR(9999); } continue;
        }
      } /* smallest cfg */
//      else { nbp[rd]+=2; } /* 38+1 but 37+1? if . not possible, u also dont work??? tJ++ */
    } /* for(;;) */
    sum=p->norm2;
    nu=tnu; nd=tnd; /* restore gobal nu, nd */
    if (fwrite(buf,Bsize,bi2,fl1)!=bi2) F0("fwrite l1");
    fclose(fl1); /* flush */
    free(buf);
    printf(" ns2.27 end found= %8lu of %8.0f srate= %5.2f sgn0= %lu ns.t= %.2f min\n",
           (ULong)i2, 1.*nt, nt*1./i2, (ULong)num0, (itime()-t1)/60.);
  } /* mpi_i=0 only, ToDo change to all mpi_i's for ofs,len */
#ifdef CONFIG_MPI
  if (nsalgo==1) {
    /* send (i2) from root and receive (i2) by all other ranks */
    MPI_Bcast(&i2,sizeof(ULong),MPI_BYTE,0,MPI_COMM_WORLD);
  }
#endif

  if (nsalgo==0) /* old pre2.27 method, parallel-scanning of scfgs */
  {
    double i3b= i3;
    if (mpi_n<20 && !mpi_i) printf(" HINT: (algo^=16) may be faster\n");
#if CM(1,0,0)  /* reduce search space by factor 1/2 for ud-sym */
    if (nu==nd && sym.num_lm==nn && sym.wud) i3/=2;
      /* ??? does not work for LM<NN */
//    if (sym.num_lm!=nn) { /* test JS 2012-05-20 */
//      tbase ib0=ll1(0);
//      if (i3>3 && next(ll1(i3/2-1))==ib0) i3=i3/2; // does not work!
//      if (!mpi_i) printf("DBG next(i3/2)=%s\n",ccfg(next(ll1(i3/2))));
//      // N=12*s1 0x00800bff 1000 0000 0000  1011 1111 1111
//    }
#endif                  
    // new 2012-06-24 we trust maxscfg with latest bug fixes
    // ToDo: problem on 2012-11-05 for CONFIG_ABEL + read 3syms + -t2
    // ToDo: better name lastsymcfg instead of maxsymcfg?
#if 0 // buggy 2012-11-05, replace get_maxscfg !!!
    i3= ll2(maxscfg0)+1;
#else
    if (!mpi_i) printf(" ToDo: fix get_maxsymcfg code to reenable speedup (Joerg2012)\n");
#endif
    if (!mpi_i) printf(" i3 reduced by factor %.2f\n",i3b*1./i3);
#ifdef CONFIG_S1SYM
    if (sym.num_lm!=nn) { 
      /* avoid a bug for (s=1,JJ,mpt), cfgs counted twice */
      /* 2012-01 test N=7 s=1 LC n1=19 -9.57219835 mpi2=OK pt2=OK noProbs */
      // N=7 mpi=8 n1=19(+1 bad) see lc_s1.gpl
#if 0
      i4=1; /* ToDo: try to remove this again, bad sum_n2! i4=pt*mpi */
#endif
    }
#endif
/*
    Better scaling diskless versions:
    - searchpackets= 1/n...1e6 store result-packets on free nodes or local
    - write to local buffer size = 1e6
      copy from buffer to new chunk size=(0...1e6)
        + repeat with cfg += threads*n
      on 1000 chunks per node? interleaved for nodes (simple resort)?
      table/list of tbase-chunks and length_of_chunk
       mit pointer vorgaenger? + nachfolger thread/position
     1. create l1buf  of length n1e=min(maxveclen/2,max_estimated_node_n1)
        and    l1buf2 of length 1M
     2. check 0..1M(@node1)..2M(@node2)... and write to local-l1buf2
     3. distribute n1found[] over nodes
     4. copy l1buf2 to l1buf
        goto 2. +nodes*1M
     5. redistribute l1buf, shorten using remalloc 

 */
    numans=i2=0; num0=0; sum=0;
    /* write to files within threads to avoid unbalanced memory */
    // rewrite so that can return earlier (after fiex time) and rest work
    //   will be distributed under the tasks? 
    /* i3=(N!)(nu!nd!) i4=num_threads=mpi*pt or 1 */
    /* new: 2...m steps, first distribute 1/16 i3, than 1/4 i3, than 1/2?
     *   distribution depends from prefered run, go from smaller to bigger
     *   steps (which a parallel computed)
     * ToDo: divide in compute chunks (extra loop of about 2..8)
     */
{   // split for better balance (ToDo)
    // for heterogen clusters? (every thread tries to stop after
    //   1 minute, 2 minute, 4minute etc. ???) more status messages
    // send different lengths to the workers? depending from 
    ULong idx_start=0, // last start index 0...i3 within chunk
          idx_stop=0,  // last end index i4 .... i3 (end_idx+1)
          l1range=i3;    // range of first run 1/256 1/16 1, renamed 2012-11-14
    int  t1=itime();   // chkpt_time
    int io=0; // outer loop index
    int mintime=999999, maxtime=0, sumtime=0; // computing time per task, summed threads

/* 2012-04 new version use mpi to transfer data to n0, n0 writes to file */
    /* truncate/reset file l1_0000.dat */
    if (!mpi_i) {
      fl1=fopen("tmp_shared/l1_0000.dat",((checkpoint)?"ab":"wb"));
      if (!fl1) F0("fopen_l1");
      // setvbuf( fl1, NULL, _IONBF , 0);
      i2= ftell(fl1)/Bsize; // long
      printf(" start_i2= %.0f Bsize=%d mpi_n=%d\n", 1.*i2, Bsize, mpi_n);
      // ToDo: full disk on big jobs!?
      // fclose(fl1);
    }
    /* big values alloc more memory but do less communication (can cause slowdown) 
     * search range per thread 1**20 causes slow down on satuated 100tasks*1GbE 
     *  sicortex.mips.5GHz 1<<20 cpu=80%(32MB,3-30s/io), 1<<25 cpu=100%(254MB,2-39m/io) tri48
     *   but cpuX=50min cpu0=33min diff=17min(MPI+NFS)
     * range=((ULong)1)<<22; // 100MB 1024 256 64 16 (2^(8*2)) start with small enough?
     * set to highest value is good idea for biggest systems but is limited by hardware 
     *  ToDo: 2*buf needed for async transfer_mpi/disk 17min vs. 33min
     */
    l1range=((ULong)1)<<25; // 254MB * tasks/cpu, start with small enough?
    l1range=((ULong)1)<<24; // 145MB * tasks/cpu, start with small enough?
    // 1<<28 = 1.8GB for 32GB/8cores but cpu=20% on 14*1GbE (dont fit to 4GB/6cores)
    //  ToDo: better fill 16MB buffer (partly) by computing 8*16M-chunks,
    //        on overflow restart with chunk/2, ggf. underflow chunk*2?
    // ToDo: testmalloc ggf l1range/2 + MPI_BCast (lowmem-nodes)
    //       may fail, malloc race condition may alloc more than allowed 2012
    if (l1range>i3/2/mpi_n) l1range=i3/2/mpi_n; // 2012-06-22
    if (l1range==0) l1range=1;   // avoid stupid error
    if (l1range>i3) l1range=i3;
    // if (l1range<i3>>16) l1range=i3>>16;  // 27er s=1 nc=2e+15 l1range=3e10 OOM
    idx_start= 0; // default start, we may use higher value for chkpt1.resume
    if (checkpoint>=1) idx_stop= idx_start= ll2(startcfg0); // 2012-06-11
    if (!mpi_i) 
      printf(" start_i1= %s %.0f\n", ccfg(ll1(idx_start)), 1.*idx_start);
    // if (!mpi_i) printf(" malloc %.0f * %d\n",l1range*1.,(int)Bsize);
    for ( ppns=&pns, ii=0; ii<pt_n; ii++ ) {
      if (!ppns) ERR(9011);                     /* check buffer for thread */
      ppns->buf=(unsigned char *)malloc(/*L1_BUFLEN 2012-04*/l1range*Bsize);
      if (!ppns->buf)
         F2("malloc %.2e L1_buf on node %d, plz reduce range", l1range*1.*Bsize, mpi_i);
      ppns->buflen= l1range; // ULong in units Bsize, available flag 2012-11-14
      if( ii<pt_n-1)
         ppns->next = (tp_ns *) malloc(sizeof(tp_ns)); // ToDo: alloc array outside loop!?
      ppns=ppns->next;
    } // start threads ii
    i4=pt_n*mpi_n; // 2012-06-26 
    /* divide into data chunks (balance compute time)*/
    for (io=0; idx_start < i3; idx_start=idx_stop, io++) {
      // stop=from would(should) be dummy call! ToDo
      if (l1range!=(l1range<<2)/4) // may be zero, avoid false alarm
         F0("l1range overflow, please report to the author");
      // ToDo: if loop is to small and enough mem enlarge l1range
      // l1range<<=2; // 1/1024 1/256 1/64 1/16 1/4 rest (2^(8*2)) growing size
      // ... but remalloc needed!
      idx_stop+=i4*l1range; // end of chunk = start of next chunk
      if (idx_stop>i3) { l1range=(i3-idx_start+i4-1)/i4; idx_stop=i3; }
      if ((!mpi_i)&&(io<1/*||(io&(io-1))==0*/)) /* io == power of 2 */
        /* 9 chars - .e+xx = 4 digits */
        printf("DBG: io= %6d start%%= %6.3f range%%= %9.5f t+= %s\n",
          io, idx_start*100./i3, l1range*100.*mpi_n*pt_n/i3, sclock(~0l));
      // ToDo min+maxtime
      for ( ppns=&pns, ii=mpi_i*pt_n; ii<i4 && ii<(mpi_i+1)*pt_n; ii++ ) {
        if (!ppns) ERR(9011);                     /* check buffer for thread */
        ppns->block=ii;                           /* block number */
        /* v2.26 (propto 2^ii)=unbalanced, try ii^2?, v2.27 (propto ii) */
//        ppns->i1=idx_start+(l1range+i4-1)/i4*(ii);   /* start of block */
//        ppns->n =idx_start+(l1range+i4-1)/i4*(ii+1); /* stop = start block+1 */
 // no overlap, but overflow? 2012-01-20
        ppns->i1=idx_start+l1range*(ii)  /*/i4*/; /* start of block */
        ppns->n =idx_start+l1range*(ii+1)/*/i4*/; /* stop = start block+1 */
        if (ppns->i1>i3) ppns->i1=i3;             /* aware of rest block */
        if (ppns->n >i3) ppns->n =i3;             /* aware of rest length */
        if (ppns->i1>i3) ppns->n = ppns->i1 = 0;  /* empty range */
        // malloc outside loop, l1range/2 in case of malloc fail (malloc-slowdown?)
        //  ppns->buf=(unsigned char *)malloc(/*L1_BUFLEN 2012-04*/l1range*Bsize); 
        //  ppns->buflen=l1range; // old: L1_BUFLEN 2012-04
        ppns->norm2=0;   ppns->numans=0; ppns->b.o=2;
        ppns->num=0;     ppns->num0=0;
        // sprintf(ppns->fname,"tmp_shared/l1_%d%04d.tmp",io,ii); /* unique name */
        /* each thread writes its file, aware of max file handles! */
        parallel_start(ii%pt_n, ns_thread, ppns);  /* ! ! ! */
        // if( ii<i4-1 && ii%pt_n<pt_n-1)
        //   ppns->next = (tp_ns *) malloc(sizeof(tp_ns)); // ToDo: alloc array outside loop!?
        ppns=ppns->next;
      } // start threads ii
{
      // new: write one file, old 33000++ files caused problems 2012-04
      //  ToDo: send all data to task0
      static ULong count_per_io=0; // debugging number found configs per printf
#ifdef CONFIG_MPI
      MPI_Status   status;
#endif
      // fl1=fopen("tmp_shared/l1_0000.dat","ab"); if(!fl1)F0("reopen_l1");
      // ToDo: send to mpi=0 thread and store full l1 there
      // if (vvv && !mpi_i) printf(" join threads\n");   /* easier to debug */
      for ( ppns=&pns,ii=mpi_i*pt_n; ii<i4 && ii<(mpi_i+1)*pt_n; ii++ ) {
        parallel_wait(ii%pt_n);
        // save_l1(ppns->l1p,ppns->l1s,i2); /* write to tmp_shared/l1_%04d.dat @pos i2 */
           /* uchar *l1p, ulong stored, ulong ofs */
        count_per_io+= ppns->num; // for debugging
        i2+=ppns->num;        sum+=ppns->norm2; 
        numans+=ppns->numans; num0+=ppns->num0;
        sumtime += ppns->time0;
        if (mintime>ppns->time0) mintime=ppns->time0;
        if (maxtime<ppns->time0) maxtime=ppns->time0;
        /* may be we should store ppns->num too for later MPI distribution */
        /* distribute ppns->num per mpi-all-to-all for mpi-code? */
#if 0 // to much output!?
        if (vvv&2 && (ii<2 || ii>=mpi_n*pt_n-2))
          printf(" node %2d.%2d n1= %9.0f n_ans= %9.4g sum_n2= %9.4g t[m]= %5.2f\n",
               mpi_i, ii%pt_n, (double)i2, (double)numans, (double)sum, ppns->time0/60.);
#endif
        // ToDo: send to task 0 which allone writes to the l1-file
#ifdef CONFIG_MPI
// if (mpi_i) printf("DBG.n%d: wait for token from 0\n",mpi_i);
// without blocking, task0 will get OOM on some MPI-implementations 2012-06 (mpich)
        if (mpi_i) { int b2; // wait for a token, see OOM-Problem on task0
          MPI_Recv(&b2,1,MPI_INT,0,1,MPI_COMM_WORLD,&status); // blocking
        }
// if (mpi_i) printf("DBG.n%d: send to 0 count=%d\n",mpi_i,(int)ppns->num);
        if (mpi_i)
          MPI_Send(ppns->buf,ppns->num*Bsize,MPI_CHAR,0,0,MPI_COMM_WORLD);
// if (mpi_i) printf("DBG.n%d: send to 0  OK\n",mpi_i);
        if (!mpi_i)
#endif
        if (ppns->num) /* ToDo: pt_n>1 untested 2012-05 */
        if (fwrite(ppns->buf,Bsize,ppns->num,fl1)
                                != ppns->num) F0("fwrite l1");
        // now free outside io compute disptribution loop 2012-06-22
        // free(ppns->buf); // 2012-4
        // p2=ppns;
        ppns=ppns->next;
        // if (ii%pt_n>0) free(p2); // free(ppns) after access!
      } // join threads ii 
      // if (fl1) fclose(fl1);
#ifdef CONFIG_MPI
// receive and write to file
if (!mpi_i)
{     int ii;
      unsigned char *buf; /* space for results (sym-configs) (r[w]) 2012-04 */
      int count;       /* size of buffer in units of Bsize (r) 2012-04 */
      // buf=(unsigned char *)malloc( l1range*Bsize);
      buf=pns.buf; // reuse compute buffer 2012-06-22
        /* should never happen but it did 2012-05-11 after 512+ loops mpi=0 */
        /*   pot. reasons: slurmd was 400MB? at n0, L1_BLen was 8*l1range */
        if (!buf) { E0("malloc");system("free;ps aux");F0("malloc"); }
                      
      for (ii=pt_n;ii<mpi_n*pt_n;ii++) { int b2=0;
// printf("DBG.n0: send token to ii=%d\n",ii);
        MPI_Send(&b2,1,MPI_INT,ii/pt_n,1,MPI_COMM_WORLD); // token to ii
        // token is needed to avoid OOM by MPI caching (n-to-1-Data-transfer) 2012-06
// printf("DBG.n0: recv from ii=%d max=%.0f\n",ii,1.*pns.buflen);
        MPI_Recv(buf,pns.buflen*Bsize,MPI_CHAR,ii/pt_n,0,MPI_COMM_WORLD,&status);
        MPI_Get_count(&status, MPI_CHAR, &count);
// printf("DBG.n0: recv from ii=%d OK count=%d\n",ii,count);
        if (fwrite(buf,Bsize,count/Bsize,fl1)
                  != (size_t)count/Bsize) F0("fwrite l1");
        count_per_io += count/Bsize;
      }
      /* lot of output (sumcount!=0) if nosym! */
      if (/*sumcount ||*/ (io<4||(io&(io-1))==0)) { /* io == power of 2 */
        /* %9.4g = 9 chars - .e+xx = 4 digits */
        printf("DBG: io= %6d stop%%= %7.3f count= %10.5g t+= %s\n",
          io, idx_stop*100./i3, 1.*count_per_io, sclock(~0l));
        count_per_io= 0;
        fflush(fl1); // better checkpointing
      }
} // (!mpi_i)
#endif /* CONFIG_MPI */
} // join results mpi+pthreads
      // ToDo: writing chkpt1 after time>=chktp_time/2
      if (checkpoint<=1 && !mpi_i)
      { int t2=itime();FILE *f1;
        extern int chkpt_time;
        if (t2-t1 >= chkpt_time/2) {
          // we need consistency, else s1cfgs may be lost 2012-06-25
          fflush(fl1); // consistent checkpointing
          fsync(fileno(fl1)); // sync file buffers to disk for consistency, unistd.h
          printf(" w.chkpt1 %s\n",ccfg(ll1(idx_stop)));
          f1=fopen(FCHKPT"1","wt");fputs(ccfg(ll1(idx_stop)),f1);fclose(f1);
          // ToDo: store i2 + idx_start, umlauf? ???
          if (checkpoint<1) { f1=fopen(FCHKPT,"wt");fputs("1\n",f1);fclose(f1); }
          checkpoint=1;
          t1=t2; // chkpt1 timer reset
        }
      }
    } // io++ idx_start+=... outer loop (l1range chunks)
    for ( ppns=&pns,ii=0; ii<pt_n; ii++ ) { // syscall outside main loop
      free(ppns->buf); /* 2012-4 */
      ppns->buflen=0;  /* 2012-11-14 not available flag */
      p2=ppns; ppns=ppns->next;
      if (ii>0) free(p2); /* free(ppns) after access! */
    } // join threads ii 
    if (!mpi_i) fclose(fl1);
#if 0 // to much output!?
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (vvv&2 && (mpi_i<2 || mpi_i>=mpi_n-2))
      printf(" node %2d n1= %9.0f n_ans= %.4g sum_n2= %9.4g t+= %s\n",
           mpi_i, 1.*i2, 1.*numans, 1.*sum, sclock(~0l));
#endif
} // split for better balance
#ifdef CONFIG_MPI
    { /* on Tru64+SLES9 MPI_UNSIGNED_LONG_LONG is not available */
      ULong lsend[4], lrecv[4]; int err; 
      lsend[0]=i2;     /* number of states counted in this MPI task */
      lsend[1]=numans; lsend[2]=sum; lsend[3]=num0;
      err=MPI_Allreduce( lsend, lrecv, 4,
                         MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
      /* ToDo: min/max for debugging/balance */
      i2=    lrecv[0]; /* overall number of counted states */
      numans=lrecv[1]; sum=   lrecv[2];
      num0=  lrecv[3]; /* number of forbidden symmetry states */ 
    }
#endif
    /* old: concat cat l1_*.tmp > l1_*.dat # replaced by seq. function */
    MPI_Barrier(MPI_COMM_WORLD); /* wait for node0 */
  } // nsalgo==0 for parallel-scanning scfgs

  if (!mpi_i) {
    printf(" ns: n1= %.15g sgn0= %.8g n_ans= %.15g sum_n2= %.15g t+= %s\n",
           1.*i2, 1.*num0, 1.*numans, 1.*sum, sclock(~0l));
    if (sum!=i3 && !mpi_i) printf(" sum!=i3");
    // get last element Little-Endian! DBG-2012-05-23
    tbase maxscfg=0; load_l1((unsigned char*)&maxscfg, 1, i2-1);
    // ToDo: error from load_l1?
    printf(" gt.maxscfg= %s from getmaxcfg()\n", ccfg(maxscfg0) );
    printf(" ns.maxscfg= %s from l1[%.0f-1]\n",  ccfg(maxscfg), i2*1.);
    if (maxscfg != maxscfg0 && !mpi_i) E0("bad maxscfg ???, ignore");
    /* write n1 to disk, does not cost much CPU */
    /* if (chkpt_mode & 1) */
    {FILE *f;f=fopen(FCHKPT"2","w");if(f)fprintf(f,"%lu",i2); // n1=i2
             else E0("w."FCHKPT"2");if(f)fclose(f);}
    {FILE *f;f=fopen(FCHKPT,"wt");fputs("2\n",f);fclose(f);}
    printf(" ns.t=%s checkpoint2",sclock(itime()-l1));
  }
  if (checkpoint<2) checkpoint=2;
  return(i2);
}

/* return a unique(?) list of counted different pairs = ising energy index
   used to block hilbertspace using model-symmetries in future versions
   uu+dd ud+du ..+33 .3+3. .u+.d+u.+d.+3u+3d+u3+d3  (16pairs -> 5sympairs)
   the list is allocated and must be freed after use
   mode|=2: output list
   mode|=1: return list (else free list, return 0)
   buf = result-array or NULL (if NULL, alloc memory)
   called by: ini_norm2 storeh2 x_out => replace by
      int get_ising_erg(b1,mode,weights)
      return product j1ud*a1+j2ud*a2+... removes need of storing array
 */
int *get_dcorr(tbase b1, int mode, int *buf) {
  int *result, i1, i2, i3, ib1, ib2;
  /* map an site pair nisnjs to a Ising index to count up */
  const char map[16]= /* fits to L1 cache line !? */
    {  2, 3, 3, 4 ,  3, 0, 1, 3 ,  3, 1, 0, 3 ,  4, 3, 3, 2 };
   /* .. .u .d .3   u. uu ud u3   d. du dd d3   3. 3u 3d 33 => 0..4 = 3bit */
   /* index: 0=uu+dd  1=ud+du  [2=..+33  3=.u+u.+.d+d.  [4=.3+3.]] */
   if (buf) result = buf;
       else result = (int *) malloc(sym.num_dcorr*CM(2,4,5)*sizeof(int));
   if (result) {
     for (i1=0;i1<sym.num_dcorr*CM(2,4,5);i1++) { result[i1]=0; } /* reset */
     for (i2=   0;i2<nn-1;i2++)    /* N*(N-1) site pairs */
     for (i3=i2+1;i3<nn  ;i3++) {
           i1=sym.corr[i2][i3];    /* access N*N pair array */
           ib1=geti(b1,i2);  /* 0=. 1=u=o 2=d=x 3=u+d=Q */
           ib2=geti(b1,i3);
           // printf("DBG: %d %d  %d %d  %d\n",i2,i3,ib1,ib2,i1);
           /* PM600MHz 32+8 isinghash=14s for 4*if(&&||)else */
           /* PM600MHz 32+8 isinghash=17s(int),12s(char) for array-access */
           result[CM(2,4,5)*i1+map[4*ib1+ib2]]++;
           // => ising+=weight[CM(2,4,5)*i1+map[4*ib1+ib2]]; /* +Ji1 for ud */
     }
     if (mode&2) {
       /* printf("\n dcorr: cfg=%s", ccfg(b1)); */
       printf(" dcorr:");  /* different pair correlations */
       for (i1=1;i1<sym.num_dcorr;i1++)
        for (i2=0; i2<CM(2,4,5)-1; i2++) /* dont print CM(uu,.u,.3) */
         printf(" %2d",result[CM(2,4,5)*i1+i2]); /* [j1,ud-.u] [j2,ud-.u] .. */
     }
   }
   /* free if temporary allocated, for output only */
   if ((!(mode&1)) && (!buf) && result) { free(result); result=NULL; }
   return result;
}

/* idea: sort l1 for norm2 (better ising-erg), to avoid storing norm2 */
#define N2Blen 1021  /* histogram size, prime number */
int     n2b[N2Blen]; /* histogram, hashed */
Tnorm2  n2h[N2Blen]; /* hash table, norm2 */

#if 0
tbase *b_l1=NULL; /* table: first_sym-config of each block (global) */
#else
tbase *mpi_l1=NULL; /* table: first_sym-config of each node (global) */
#endif

extern ULong x_max; /* ToDo: maximum number output xval.h => debug.h */

/* ----- initialize symmetric base (cfg-tables l1) v1.53 ----- 
 * calculate orbit length or norm of each sym-config
 *  also do some speed tests and test new ideas
 * ToDo: should be done with creation of l1
 * v2.33 norm2[] table removed to simplify code, less MEM, little more CPU
 */
void ini_norm2(void){
  ULong i1;
#if 0  // Aug2010 not used anymore?
  Tnorm2 sum2=0l,n2max=0; int n2hash, n2num=0;
#endif
  int i2;
  tbase ib1;
  static struct sbase b1 /* b2=tmp */;
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (vvv && !mpi_i) printf("\n ini_norm2");
  b1.o=0;

  ERR(-1); /* reset errno for ERR(600) */
  /* initialize hash table (for norm2 distribution) */
  for (i1=0l; i1<N2Blen; i1++) n2h[i1]=n2b[i1]=0;

  /* create scfg-range per block table */
  if (vvv&2 && !mpi_i) printf(" ini_mpi_l1[]");
  if (!mpi_l1) mpi_l1=(tbase *)malloc(sizeof(tbase)*mpi_n);
  if (!mpi_l1) ERR(9011);
  for (i2=0;i2<mpi_n;i2++) mpi_l1[i2] = 0; /* clear for debugging */
  /* mpi_l1 must be set also for len=0 */
  mpi_l1[mpi_i] =
       ((b_ofs[pt_n])? getl1(b_ofs[0]) /* first config of this node */
                     : ~(tbase)0 ); /* dummy cfg > max_cfg */
  // b_l1[mpi_i*pt_n+pt_n]= getl1(node_n1-1); /* ??? */
  /* MPI_Allgather(sbuf, 100, MPI_INT, rbuf, 100, MPI_INT, comm);
   *     AllGather: rbuf[  0.. 99]=proc0.sbuf[0..99]
   *                rbuf[100..199]=proc1.sbuf[0..99]
   *                ...           =procn.sbuf[0..99]
   */
  /* can we use a part of the receiverarray as sendarray? yes v2.27 */
#ifdef CONFIG_MPI
{
  tbase tmp_l1=mpi_l1[mpi_i]; /* sendbuf != recvbuf, v2.36 */
#if 0
  if (vvv&4) {
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (mpi_i<4 || mpi_i>=mpi_n-4)
      printf("\n DBGpreAG  node= %2d mpi_l1= %s",mpi_i,ccfg(mpi_l1[mpi_i]));
    else if (mpi_i==4) printf("\n ...");
  }
#endif
  MPI_Allgather(&tmp_l1, sizeof(tbase), MPI_BYTE,
                 mpi_l1, sizeof(tbase), MPI_BYTE, MPI_COMM_WORLD);
#if 0
  if (mpi_i==mpi_n-1 && vvv&4)
  for(i2=0;i2<mpi_n;i2++) 
    if (i2<4 || i2>=mpi_n-4)
      printf("\n DBGpostAG node= %2d mpi_l1= %s", i2, ccfg(mpi_l1[i2]));
    else if (i2==4) printf("\n ...");
#endif
}
#endif

  if (vvv&2) { /* speed test, see scfg2idx() for results */
    ULong j=~0, loops; long t2=0, t1=itime();
    tbase ib2=0; /* JS2012-01 do a simple check that i1<next(i1) */
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (!mpi_i) printf(" scfg2idx_speedtest.start\n");
    for (loops=0;t2<1 && loops<100000;loops++) { /* run until 1s gone */
      for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
       ib1=getl1((ULong)i1);       /* l1 is the node part of full l1 */
       if (ib2>=ib1 && i1!=0) break;
       j=scfg2idx(ib1); if (j!=i1) break;
       ib2=ib1;
      }
      if (i1<node_n1) { // JS2012-01-20
        printf("ERR task%d (j=%lu)!=(i1=%lu) or ib2>=ib1 ll2=%lu\n",
          mpi_i, j, i1, ll2(ib1));
        printf("ERR ib2 %s\n",ccfg(ib2));
        printf("ERR ib1 %s\n",ccfg(ib1));
        printf("ERR j-1 %s\n",ccfg(getl1(j-1)));
        printf("ERR j   %s\n",ccfg(getl1(j)));
        printf("ERR j+1 %s\n",ccfg(getl1(j+1)));
        printf("ERR j+2 %s\n",ccfg(getl1(j+2)));
        sleep(10); // give other tasks a chance to show further errors
        F0("inkonsistent l1")
      }
      t2=itime()-t1;
    }
    if (!mpi_i)
      printf(" scfg2idx_speedtest.end t= %s loops= %lu\n", sclock(t2), loops);
  }
  
  if (vvv&4) /* not needed at the moment */
  { /* create Isingtable for blocking criteria, Idea from Oct07 */
    /* ToDo: also output a num_states_per_IsingErg I1-I2-matrix */
    int j1,                    /* loop-vars */ 
        min0, max0,             /* min/max numPairs */
        ising,                  /* ising erg, may be exotic hash of all Js */ 
        *buf=NULL;              /* buffer for counted pairs (IsingErg) */
#define MaxIsing (B_NUM)    /* should be prime but 4*(pt_n)..4*NN */
    long numIsing[MaxIsing];   /* IsingErg[j1,j2,..] for load balance */
    long minI,maxI,t1=itime();

    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (!mpi_i)
      printf("\n ini_norm2.IsingBlocks:");
    min0=MaxIsing; max0=0;
    for(j1=0;j1<MaxIsing;j1++) numIsing[j1]=0; /* init hash array */
    /* ToDo: hash should leave near J1-Ising-ergs at near hashs */
    if (sym.num_dcorr*CM(2,4,5)>22)
    for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
      ib1=getl1((ULong)i1);
      buf=get_dcorr(ib1,1,buf); /* get dcorr-table, count ud-pairs */
      /* which bonds we take? worst case is spin glas -> add different bonds */
      /* Idea: make a monte carlo test to find a good function */
      /* IsingErg can be arteficcial build from dcorr, ex: j1+j4/2+... */
      /* ToDo: do hashing within get_dcorr? => hashed j1-jmax-Ising erg */
//    ising=((731*buf[ 5]+23*buf[15]+171*buf[25]+3*buf[35]+17*buf[45])/2)%pt_n;
      ising=((9551*buf[ 2]+23*buf[ 6]
             +1327*buf[10]+ 3*buf[14]
             +  17*buf[18]+ 7*buf[22])/2)%pt_n;
      if (min0>ising) min0=ising;
      if (max0<ising) max0=ising;
      if (ising>=MaxIsing) ising=MaxIsing-1;
      /* hash indicees to avoid overflow */
      numIsing[ising]++;
    }
    if (!mpi_i)
      printf("\n Ising range for node= %2d: %d-%d", mpi_i, min0, max0);
    /* output Ising distribution, ToDo: only min, max, variance? */
    minI=n1; maxI=0;
    for(j1=min0;j1<MaxIsing && j1<=max0;j1++){
      if (!mpi_i)
        printf("\n node= %2d Ising hash=%2d: %6ld", mpi_i, j1, numIsing[j1]);
      if (numIsing[j1]<minI) minI= numIsing[j1];
      if (numIsing[j1]>maxI) maxI= numIsing[j1];
    }
    if (!mpi_i)
      printf("\n node= %2d Ising block size: %lu-%lu-%lu (%.0f%%-%.0f%%)"
             "\n ising.t=%s",
            mpi_i, minI, node_n1/pt_n, maxI,
            minI*100./(node_n1/pt_n),
            maxI*100./(node_n1/pt_n), sclock(itime()-t1));
    if (buf) free(buf); /* malloc by get_dcorr() */
  }
  /* ToDo: sum hash tables? + pthread */
  MPI_Barrier(MPI_COMM_WORLD); /* debug */

#if 1
  if (vvv && !mpi_i)
    printf(" create n2hashtables n2b,n2h - skipped (ToDo)\n");
#else /* slow! */
  if (vvv && !mpi_i)
    printf(" create n2hashtables n2b,n2h - (ToDo parallel)\n");
  /* ToDo: remove n2, no significant speedup and MPI-Probs */
  /* if (norm2) ToDo: ini sqrt[n2a][n2b]-table when n2a!=n2b often? */
  for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
   ib1=getl1((ULong)i1);       /* l1 is the node part of full l1 */

   b_ifsmallest3(ib1,&b1); /* called only to get norm2 */
   sum2+=b1.norm2; /* sum2 == numconf ! */
   if(b1.norm2 > n2max) n2max=b1.norm2;	/* store maximum orbit length */

   n2hash=(int)(((ULong)b1.norm2)%N2Blen);
   if (!n2b[n2hash]) { n2b[n2hash]++; n2h[n2hash]=b1.norm2; n2num++; }
              else if (n2h[n2hash]==b1.norm2) n2b[n2hash]++;
              else { n2b[0]++; /* collision */}
  }
  if (err_fatal==600) ERR(-1); /* reset error 600 */
  if (vvv&4) {
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (!mpi_i) printf("\n orbit lengths (norm2): ");
    for(i1=1; i1<N2Blen; i1++) if (n2b[i1]) {
      if (mpi_i<4 || mpi_i>=mpi_n-4)
      printf("\n node %2d norm2= %15.10g %8d times",
              mpi_i,(double)n2h[i1],n2b[i1]);
      else if (mpi_i==4) printf("\n ...");
    }
    if (n2b[0]) printf("\n hash_collisions= %8d (should be zero)",n2b[0]);
  }

  if (vvv&4) {
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (mpi_i<4 || mpi_i>=mpi_n-4)
    printf("\n node %2d n2num= %d  n2max= %.10g sum_norm2= %.15g t+= %s",
            mpi_i, n2num, (double)n2max, (double)sum2, sclock(~0l));
    else if (mpi_i==4) printf("\n ...");
  }

  /* does not work for Tnorm2 == double */
  if ( ((ULong)n2max)>>(sizeof(ATnorm2)/2*8) == 0 && !mpi_i)
    printf("\n# spare memory: Change ATnorm2 to short or char.\n");

  MPI_Barrier(MPI_COMM_WORLD); /* debug */
#endif

#ifndef CONFIG_MPI
  /* MaxSBase set to 1 to spare Memory but b_getbase will fail v2.33 */
  if ((vvv&128) && !mpi_i) {
    printf("\nini_norm2 vvv&128 (node=0 only):");
    for(i1=0l;i1<node_n1 && i1<x_max;i1++){
      /* show base for debugging purpose */
      b_ifsmallest3(getl1((ULong)i1),&b1); /* different results for S>=1 ??? */
      printf("\n %6d",(int)b1.norm2);
 /*     b_smallest(getl1((ULong)i1),&b1); *
  *     printf(  " %6d",(int)b1.norm2);   */
      b_getbase(getl1((ULong)i1),&b1);
      printf(  " %6d |%.15g>=",(int)b1.norm2,(double)i1);
      for(i2=0l;i2<b1.num;i2++) 
        printf("\n %32s*%s",outphase(b1.is[i2]),ccfg(b1.ib[i2]));
    }
  }
#endif
  if (vvv && !mpi_i) printf(" ini_norm2.end\n");
  return;
}


/* last modifications: 
   1998-148 type tbase j,m substituted by ulong etc.
   Dez98: general tbase-type independent scfg2idx-fkt. + better errorhandling
 */

/* convert base to index, this is good break point for debugging */
ULong scfg2idx_err(tbase cfg){	/* called if error in scfg2idx() occurs */
  ULong j, m;
  static int err=0;
  static ULong m2=1l;    /* not thread safe */
  static struct sbase b; /* static makes it 100x faster, only used if error */
  int node, i1, i2, iu, id; 
  err++;
  if (err>128 && 0!=(err&(err-1))) return ~0; /* limit error msgs lg2 */
  printf("\nscfg2idx_err %s not found (mpi_i=%d, numerr=%d)",
          ccfg(cfg), mpi_i, err);
  /* if getl1 was called with index_out_of_range, we have wrong bits */
  for (iu=id=i1=0;i1<nn;i1++) { i2=geti(cfg,i1); if(i2&1)iu++; if(i2&2)id++; }
  if (iu!=nu || id!=nd) {
    printf("\nscfg2idx_err %d: scfg has nu,nd= %d %d which is wrong",
           mpi_i, iu, id);
    ERR(9999);
  }
  node=scfg2node(cfg);
  if (node!=mpi_i) {
    printf("\nscfg2idx_err %d: scfg belongs to wrong node %d",
              mpi_i, node);
    ERR(9999);
  }
  /* recompute the index */
  while ((m2<<1)<node_n1) m2<<=1;	/* Feb00+Nov07 */
  for(j=0l,m=m2;m;m>>=1){if((j|m)<node_n1)if( (getl1(j|m)<=cfg) )j|=m;}

  if (!(getl1(j)==cfg)) /* ok, its still not ok */
  {			/* not found */
   b.o=0; /* getbase may fail on size optimized code without panic */
   b_getbase( cfg,&b);printf("\n     cfg_gb= %s phase= %s",ccfg(b.scfg),outphase(b.sign));
   b_smallest(cfg,&b);printf("\n     cfg_sm= %s phase= %s",ccfg(b.scfg),outphase(b.sign));
   if (b.sign==0) printf("\n err sign==0 %s numerr=%d",ccfg(b.scfg),err);
   else { // ToDo: temporary test (remove it)
     if (cfg!=b.scfg) { cfg=b.scfg;
       b_smallest(cfg,&b);printf("\n    cfg_sm2= %s phase= %s",ccfg(b.scfg),outphase(b.sign));}
     if (cfg!=b.scfg) { cfg=b.scfg;
       b_smallest(cfg,&b);printf("\n    cfg_sm2= %s phase= %s",ccfg(b.scfg),outphase(b.sign));}
     if (cfg!=b.scfg) { cfg=b.scfg;
       b_smallest(cfg,&b);printf("\n    cfg_sm2= %s phase= %s",ccfg(b.scfg),outphase(b.sign));}
   }
   cfg=b.scfg;
   for(j=0l,m=m2;m;m>>=1){ if((j|m)<node_n1)if((getl1(j|m)<=cfg))j|=m; }
   if(!(getl1(j)==cfg)){       /* not found again !!! */
     /* if (err<3) */ {
       if(j> (ULong)0) printf("\n l1[%6lu]= %s",0l ,ccfg(getl1( 0l)));
       if(j> (ULong)1) printf("\n l1[%6lu]= %s",1l ,ccfg(getl1( 1l)));
       if(j-2<node_n1) printf("\n l1[%6lu]= %s",j-2,ccfg(getl1(j-2)));
       if(j-1<node_n1) printf("\n l1[%6lu]= %s",j-1,ccfg(getl1(j-1)));
       if(j  <node_n1) printf("\n l1[%6lu]= %s",j  ,ccfg(getl1(j  )));
       if(j+1<node_n1) printf("\n l1[%6lu]= %s",j+1,ccfg(getl1(j+1)));
       j=node_n1-1;
       if(node_n1 > 0) printf("\n l1[%6lu]= %s\n",j  ,ccfg(getl1(j  )));
       ERR(400); /* not found, try again */
     }
     for(j=0;j<node_n1;j++) if(getl1(j)==cfg) break; /* l1 not sorted? */
     if(j==node_n1)if(err<10){ ERR(410); err_fatal++; }
   }
  }
  if (getl1(j)==cfg && err<5)
    printf("\n node %2d: now scfg2idx() is ok and returns %lu", mpi_i, (unsigned long)j);
#ifndef CONFIG_ANSATZ
  if (getl1(j)!=cfg) ERR(9999); /* 2012-11-05 this is fatal! report it! */
#endif
  return (j);
}

/*
  scfg2idx: sym-config to index within node (0..node_n1-1) (old: b2i)
  the routine is short as possible to fit into the cache
   input is spin-configuration (representator)
   output is index to tables
   ToDo:
    - convert more than one index within one call  for better cache performance
    - take 8 highest 8bit as index best_speedup=lg2(n1)/(lg2(n1)-8)=1.36
      no effect for high Sz, because all configs start with 0-bits?
    - l1_rough_idx[0..1023] Oct07 v2.27
      scfg2idx-speedtest getl1+scfg2idx+if nud=33,7  P4-1.3GHz L2=512kB
       search full table           92loops/9s (bits), 94loops/9s (interval)
       search through rough table 147loops/9s
       no search                 4900loops/9s
       lg2(n1=116769)=17bits lg2(Nrough=1024)=10bits speedup=50%
       no speedup for lg2(n1=1e9)=30bits (speedup=30%)
    Idea: reduce rough table for nodes and threads (no mem collisions!)
          as a compromise
    Idea: sort-as-searchtree: 1..10 to { 5,    2, 7,     1,3, 6,8,  ...}
                                         root  2elements 4elements 8elements
           for better caching? not really -> need hardware-search (or hash?)
    Idea: hash table? or IsingErg
      return -i if cfg outside block? -> no, use extra chck+fkt
    ToDo: search thread_l1_block only?, rename to scfg2i
                                
 */
ULong scfg2idx(tbase cfg){ /* (b=base_config=scfg)-to-index < node_n1, rename to scfg2i! */
  ULong j,ja,jb;
  ja=0; j=jb=node_n1-1;
  /* ToDo: block.h b_idx2blk() looks more nice, without +-1 */
  for (;ja<jb;) { j=(ja+jb+1)/2; if (getl1(j)<=cfg) ja=j; else j=jb=j-1; }
  /* price: log2(node_n1)*(2cmp+2add+1shft+1mem+2jmp) */
  if (getl1(j)==cfg) return j;  /* getl1 returns the node_index=0..node_n1-1 */
  return scfg2idx_err(cfg); /* slow error-routine outside the cpu-i-cache */
}

/* scfg2node: get global mpi node number from sym-config
 */
int scfg2node(tbase cfg){ /* get global mpi node number from sym-config */
  int j,ja,jb;
  ja=0; j=jb=mpi_n-1;
  for (;ja<jb;) { j=(ja+jb+1)/2; if (mpi_l1[j]<=cfg) ja=j; else j=jb=j-1; }
  /* price: log2(mpi_n)*(2cmp+2add+1shft+1mem+2jmp) */
  return j;  /* getl1 returns the node_index=0..node_n1-1 */
}
